# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import csv
from warnings import warn
from itertools import chain
from collections import OrderedDict
from contextlib import contextmanager
import logging

import numpy as np

from mvc.models.properties import (
    SignalProperty, LabeledProperty, SignalMixin, StringProperty,
    BoolProperty, IntegerProperty
)
logger = logging.getLogger(__name__)

from pyxrd.data import settings

from pyxrd.generic.io import storables, Storable
from pyxrd.generic.models import DataModel

from pyxrd.refinement.refinement import Refinement

from pyxrd.calculations.data_objects import MixtureData
from pyxrd.phases.models import Phase

from .optimizers import Optimizer

@storables.register()
class Mixture(DataModel, Storable):
    """
        The base model for optimization and refinement of calculated data
        and experimental data. This is the main model you want to interact with,
        lower-level classes' functionality (mainly 
        :class:`~pyxrd.mixture.models.optimizers.Optimizer` and 
        :class:`~pyxrd.mixture.models.refiner.Refinement`) are integrated into this
        class. 
        
        The Mixture is responsible for managing the phases and specimens lists
        and combination-matrix and for delegating any optimization, calculation
        or refinement calls to the appropriate helper class.
    """
    # MODEL INTEL:
    class Meta(DataModel.Meta):
        store_id = "Mixture"

    _data_object = None
    @property
    def data_object(self):
        self._data_object.parsed = False
        self._data_object.optimized = False
        self._data_object.calculated = False
        self._data_object.specimens = [None] * len(self.specimens)
                
        self._data_object.n = len(self.specimens)
        self._data_object.m = len(self.phases)
               
        self._data_object.scales_mask = np.ones_like(self.scales)
        if self.auto_scales:
            self._data_object.scales_mask = np.ones_like(self.bgshifts)
        else:
            self._data_object.scales_mask = np.zeros_like(self.bgshifts)
            
        if self.auto_bg:
            self._data_object.bgshifts_mask = np.ones_like(self.bgshifts)
        else:
            self._data_object.bgshifts_mask = np.zeros_like(self.bgshifts)
        
        for i, specimen in enumerate(self.specimens):
            if specimen is not None:
                data_object = specimen.data_object
                data_object.phases = [[None] * self._data_object.m for _ in data_object.z_list]
                for z_index in range(len(specimen.get_z_list())):
                    for phase_index in range(self.phase_matrix.shape[1]):
                        data_object.phases[z_index][phase_index] = self.get_phase_data_object(i, z_index, phase_index)
                self._data_object.specimens[i] = data_object
            else:
                self._data_object.specimens[i] = None
        return self._data_object

    def get_phase_data_object(self, specimen_index, z_index, phase_index):
        phase = self.phase_matrix[specimen_index, ...].flatten()[phase_index]
        return phase.data_object if phase is not None else None

    project = property(DataModel.parent.fget, DataModel.parent.fset)

    # SIGNALS:
    #: Signal, emitted when the # of phases or specimens changes
    needs_reset = SignalProperty()
    #: Signal, emitted when the patterns of the specimens need an update
    needs_update = SignalProperty()

    #: The name of this Mixture
    name = StringProperty(
        default="", text="Name",
        visible=True, persistent=True, tabular=True,
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )

    #: Flag, True if the mixture will automatically adjust phase fractions and scales
    auto_run = BoolProperty(
        default=False, text="Auto Run",
        visible=True, persistent=True, tabular=True
    )

    #: Flag, True if the mixture is allowed to also update the background level
    auto_bg = BoolProperty(
        default=False, text="Auto Bg",
        visible=True, persistent=True, tabular=True
    )

    #: Flag, True if the mixture is allowed to also update the background level
    auto_scales = BoolProperty(
        default=True, text="Auto Scales",
        visible=True, persistent=True, tabular=True
    )

    #: The tree of refinable properties
    @LabeledProperty(
        default=None, text="",
        visible=True, persistent=False, tabular=True,
        data_type=object,
    )
    def refinables(self):
        return self.refinement.refinables

    #: An integer describing which method to use for the refinement (see
    #: mixture.models.methods.get_all_refine_methods)
    @IntegerProperty(
        default=0, text="Refinement method index",
        visible=False, persistent=True,
    )
    def refine_method_index(self):
        return self.refinement.refine_method_index

    #: A dict containing the current refinement options
    @LabeledProperty(
        default=None, text="Current refinement method options",
        visible=False, persistent=True, tabular=False,
        data_type=object, store_private="all_refine_options"
    )
    def refine_options(self):
        return  self.refinement.refine_options

    #: A dict containing all refinement options
    @LabeledProperty(
        default=None, text="All refinement methods options",
        visible=False, persistent=False, tabular=False,
        data_type=object
    )
    def all_refine_options(self):
        return self.refinement.all_refine_options

    # Lists and matrices:
    #: A 2D numpy object array containing the combination matrix
    phase_matrix = None
    #: The list of specimen objects
    specimens = None
    #: The list of phase names
    phases = None

    @property
    def scales(self):
        """ A list of floats containing the absolute scales for the calculated patterns """
        return self._data_object.scales
    @scales.setter
    def scales(self, value):
        self._data_object.scales = value

    @property
    def bgshifts(self):
        """ A list of background shifts for the calculated patterns """
        return self._data_object.bgshifts
    @bgshifts.setter
    def bgshifts(self, value):
        self._data_object.bgshifts = value

    @property
    def fractions(self):
        """ A list of phase fractions for this mixture """
        return self._data_object.fractions
    @fractions.setter
    def fractions(self, value):
        self._data_object.fractions = value

    @property
    def fractions_mask(self):
        """ A mask indicating which fractions are to be optimized """
        return self._data_object.fractions_mask
    @fractions_mask.setter
    def fractions_mask(self, value):
        self._data_object.fractions_mask = value

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        """
            Constructor takes any of its properties as a keyword argument.
            It also support two UUID list keyword arguments:
                - phase_uuids: a list of UUID's for the phases in the mixture
                - specimen_uuids: a list of UUID's for the specimens in the mixture
            These should be *instead* of the phases and specimens keywords.
            
            In addition to the above, the constructor still supports the 
            following deprecated keywords, mapping to a current keyword:
                - phase_indeces: a list of project indices for the phases in the mixture
                - specimen_indeces: a list of project indices for the specimens in the mixture
                
            Any other arguments or keywords are passed to the base class.
        """

        my_kwargs = self.pop_kwargs(kwargs,
            "data_name", "phase_uuids", "phase_indeces", "specimen_uuids",
            "specimen_indeces", "data_phases", "data_scales", "data_bgshifts",
            "data_fractions", "refine_method", "data_refine_method",
            "fractions", "fractions_mask", "bgshifts", "scales", "phases",
            *[prop.label for prop in Mixture.Meta.get_local_persistent_properties()]
        )
        super(Mixture, self).__init__(*args, **kwargs)
        kwargs = my_kwargs

        with self.data_changed.hold():

            self._data_object = MixtureData()

            self.name = self.get_kwarg(kwargs, "New Mixture", "name", "data_name")
            self.auto_run = self.get_kwarg(kwargs, False, "auto_run")
            self.auto_bg = self.get_kwarg(kwargs, True, "auto_bg")
            self.auto_scales = self.get_kwarg(kwargs, True, "auto_scales")

            # 2D matrix, rows match specimens, columns match mixture 'phases'; contains the actual phase objects
            phase_uuids = self.get_kwarg(kwargs, None, "phase_uuids")
            phase_indeces = self.get_kwarg(kwargs, None, "phase_indeces")
            if phase_uuids is not None:
                self.phase_matrix = np.array([[type(type(self)).object_pool.get_object(uuid) if uuid else None for uuid in row] for row in phase_uuids], dtype=np.object_)
            elif phase_indeces and self.parent is not None:
                warn("The use of object indices is deprecated since version 0.4. Please switch to using object UUIDs.", DeprecationWarning)
                self.phase_matrix = np.array([[self.parent.phases[index] if index != -1 else None for index in row] for row in phase_indeces], dtype=np.object_)
            else:
                self.phase_matrix = np.empty(shape=(0, 0), dtype=np.object_)

            # list with actual specimens, indexes match with rows in phase_matrix
            specimen_uuids = self.get_kwarg(kwargs, None, "specimen_uuids")
            specimen_indeces = self.get_kwarg(kwargs, None, "specimen_indeces")
            if specimen_uuids:
                self.specimens = [type(type(self)).object_pool.get_object(uuid) if uuid else None for uuid in specimen_uuids]
            elif specimen_indeces and self.parent is not None:
                warn("The use of object indices is deprecated since version 0.4. Please switch to using object UUIDs.", DeprecationWarning)
                self.specimens = [self.parent.specimens[index] if index != -1 else None for index in specimen_indeces]
            else:
                self.specimens = list()

            # list with mixture phase names, indexes match with cols in phase_matrix
            self.phases = self.get_kwarg(kwargs, list(), "phases", "data_phases")

            # list with scale values, indexes match with rows in phase_matrix (= specimens)
            self.scales = np.asarray(self.get_kwarg(kwargs, [1.0] * len(self.specimens), "scales", "data_scales"))
            # list with specimen background shift values, indexes match with rows in phase_matrix (=specimens)
            self.bgshifts = np.asarray(self.get_kwarg(kwargs, [0.0] * len(self.specimens), "bgshifts", "data_bgshifts"))
            # list with phase fractions, indexes match with cols in phase_matrix (=phases)
            self.fractions = np.asarray(self.get_kwarg(kwargs, [0.0] * len(self.phases), "fractions", "data_fractions"))
            # list with phase fractions mask, indexes match with cols in phase_matrix (=phases)
            self.fractions_mask = np.asarray(self.get_kwarg(kwargs, [1] * len(self.phases), "fractions_mask"))

            # sanity check:
            n, m = self.phase_matrix.shape if self.phase_matrix.ndim == 2 else (0, 0)
            if len(self.scales) != n or len(self.specimens) != n or len(self.bgshifts) != n:
                raise IndexError("Shape mismatch: scales (%d), background shifts (%d) or specimens (%d) list lengths do not match with row count (%d) of phase matrix" % (len(self.scales), len(self.specimens), len(self.bgshifts), n))
            if len(self.phases) != m or len(self.fractions) != m:
                raise IndexError("Shape mismatch: fractions (%s) or phases (%d) lists do not match with column count of phase matrix (%d)" % (len(self.fractions), len(self.phases), m))

            self._observe_specimens()
            self._observe_phases()

            self.optimizer = Optimizer(parent=self)
            self.refinement = Refinement(
                refine_method_index=self.get_kwarg(kwargs, 0, "refine_method_index", "refine_method", "data_refine_method"),
                refine_options=self.get_kwarg(kwargs, dict(), "refine_options"),
                parent=self)

            self.update()

            self.observe_model(self)

            pass # end hold data_changed

    # ------------------------------------------------------------
    #      Notifications of observable properties
    # ------------------------------------------------------------
    @DataModel.observe("removed", signal=True)
    def notify_removed(self, model, prop_name, info):
        if model == self:
            self._relieve_phases()
            self._relieve_specimens()

    @DataModel.observe("needs_update", signal=True)
    def notify_needs_update(self, model, prop_name, info):
        with self.data_changed.hold():
            self.update()

    @DataModel.observe("data_changed", signal=True)
    def notify_data_changed(self, model, prop_name, info):
        if not model == self and not (
            info.arg == "based_on" and model.based_on is not None and
            model.based_on in self.phase_matrix):
                self.needs_update.emit()

    @DataModel.observe("visuals_changed", signal=True)
    def notify_visuals_changed(self, model, prop_name, info):
        if isinstance(model, Phase) and \
           not (info.arg == "based_on" and model.based_on is not None and model.based_on in self.phase_matrix):
            for i, specimen in enumerate(self.specimens):
                if specimen is not None:
                    specimen.update_visuals(self.phase_matrix[i, :])

    # ------------------------------------------------------------
    #      Input/Output stuff
    # ------------------------------------------------------------
    def json_properties(self):
        self.refinement.update_refinement_treestore()
        retval = Storable.json_properties(self)

        retval["phase_uuids"] = [[item.uuid if item else "" for item in row] for row in map(list, self.phase_matrix)]
        retval["specimen_uuids"] = [specimen.uuid if specimen else "" for specimen in self.specimens]
        retval["phases"] = self.phases
        retval["fractions"] = self.fractions.tolist()
        retval["fractions_mask"] = self.fractions_mask.tolist()
        retval["bgshifts"] = self.bgshifts.tolist()
        retval["scales"] = self.scales.tolist()

        return retval

    @staticmethod
    def from_json(**kwargs):
        # Remove this deprecated kwarg:
        if "refinables" in kwargs:
            del kwargs["refinables"]
        return Mixture(**kwargs)

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def unset_phase(self, phase):
        """ Clears a phase slot in the phase matrix """
        with self.needs_update.hold_and_emit():
            with self.data_changed.hold():
                shape = self.phase_matrix.shape
                with self._relieve_and_observe_phases():
                    for i in range(shape[0]):
                        for j in range(shape[1]):
                            if self.phase_matrix[i, j] == phase:
                                self.phase_matrix[i, j] = None
                self.refinement.update_refinement_treestore()

    def unset_specimen(self, specimen):
        """ Clears a specimen slot in the specimen list """
        with self.needs_update.hold_and_emit():
            with self.data_changed.hold():
                with self._relieve_and_observe_specimens():
                    for i, spec in enumerate(self.specimens):
                        if spec == specimen:
                            self.specimens[i] = None

    def get_phase(self, specimen_slot, phase_slot):
        """Returns the phase at the given slot positions or None if not set"""
        return self.phase_matrix[specimen_slot, phase_slot]

    def set_phase(self, specimen_slot, phase_slot, phase):
        """Sets the phase at the given slot positions"""
        if self.parent is not None: #no parent = no valid phases
            with self.needs_update.hold_and_emit():
                with self.data_changed.hold():
                    with self._relieve_and_observe_phases():
                        if phase is not None and not phase in self.parent.phases:
                            raise RuntimeError("Cannot add a phase to a Mixture which is not inside the project!")
                        self.phase_matrix[specimen_slot, phase_slot] = phase
                    self.refinement.update_refinement_treestore()

    def get_specimen(self, specimen_slot):
        """Returns the specimen at the given slot position or None if not set"""
        return self.specimens[specimen_slot]

    def set_specimen(self, specimen_slot, specimen):
        """Sets the specimen at the given slot position"""
        if self.parent is not None: #no parent = no valid specimens
            with self.needs_update.hold_and_emit():
                with self.data_changed.hold():
                    with self._relieve_and_observe_specimens():
                        if specimen is not None and not specimen in self.parent.specimens:
                            raise RuntimeError("Cannot add a specimen to a Mixture which is not inside the project!")
                        self.specimens[specimen_slot] = specimen

    @contextmanager
    def _relieve_and_observe_specimens(self):
        self._relieve_specimens()
        yield
        self._observe_specimens()

    def _observe_specimens(self):
        """ Starts observing specimens in the specimens list"""
        for specimen in self.specimens:
            if specimen is not None:
                self.observe_model(specimen)

    def _relieve_specimens(self):
        """ Relieves specimens observer calls """
        for specimen in self.specimens:
            if specimen is not None:
                self.relieve_model(specimen)

    @contextmanager
    def _relieve_and_observe_phases(self):
        self._relieve_phases()
        yield
        self._observe_phases()

    def _observe_phases(self):
        """ Starts observing phases in the phase matrix"""
        for phase in self.phase_matrix.flat:
            if phase is not None:
                self.observe_model(phase)

    def _relieve_phases(self):
        """ Relieves phase observer calls """
        for phase in self.phase_matrix.flat:
            if phase is not None:
                self.relieve_model(phase)

    def add_phase_slot(self, phase_name, fraction):
        """ Adds a new phase column to the phase matrix """
        with self.needs_update.hold_and_emit():
            with self.data_changed.hold():
                self.phases.append(phase_name)
                self.fractions = np.append(self.fractions, fraction)
                self.fractions_mask = np.append(self.fractions_mask, 1)
                n, m = self.phase_matrix.shape if self.phase_matrix.ndim == 2 else (0, 0)
                if self.phase_matrix.size == 0:
                    self.phase_matrix = np.resize(self.phase_matrix.copy(), (n, m + 1))
                    self.phase_matrix[:] = None
                else:
                    self.phase_matrix = np.concatenate([self.phase_matrix.copy(), [[None]] * n ], axis=1)
                    self.phase_matrix[:, m] = None
                self.refinement.update_refinement_treestore()
        return m

    def del_phase_slot(self, phase_slot):
        """ Deletes a phase column using its index """
        with self.needs_update.hold_and_emit():
            with self.data_changed.hold():
                with self._relieve_and_observe_phases():
                    # Remove the corresponding phase name, fraction & references:
                    del self.phases[phase_slot]
                    self.fractions = np.delete(self.fractions, phase_slot)
                    self.fractions_mask = np.delete(self.fractions_mask, phase_slot)
                    self.phase_matrix = np.delete(self.phase_matrix, phase_slot, axis=1)
                # Update our refinement tree store to reflect current state
                self.refinement.update_refinement_treestore()
        # Inform any interested party they need to update their representation
        self.needs_reset.emit()

    def del_phase_slot_by_name(self, phase_name):
        """ Deletes a phase slot using its name """
        self.del_phase_slot(self.phases.index(phase_name))

    def add_specimen_slot(self, specimen, scale, bgs):
        """ Adds a new specimen to the phase matrix (a row) and specimen list """
        with self.needs_update.hold_and_emit():
            with self.data_changed.hold():
                self.specimens.append(None)
                self.scales = np.append(self.scales, scale)
                self.bgshifts = np.append(self.bgshifts, bgs)
                n, m = self.phase_matrix.shape if self.phase_matrix.ndim == 2 else (0, 0)
                if self.phase_matrix.size == 0:
                    self.phase_matrix = np.resize(self.phase_matrix.copy(), (n + 1, m))
                    self.phase_matrix[:] = None
                else:
                    self.phase_matrix = np.concatenate([self.phase_matrix.copy(), [[None] * m] ], axis=0)
                    self.phase_matrix[n, :] = None
                if specimen is not None:
                    self.set_specimen(n, specimen)
        return n

    def del_specimen_slot(self, specimen_slot):
        """ Deletes a specimen slot using its slot index """
        with self.needs_update.hold_and_emit():
            with self.data_changed.hold():
                # Remove the corresponding specimen name, scale, bg-shift & phases:
                with self._relieve_and_observe_specimens():
                    del self.specimens[specimen_slot]
                    self.scales = np.delete(self.scales, specimen_slot)
                    self.bgshifts = np.delete(self.bgshifts, specimen_slot)
                    self.phase_matrix = np.delete(self.phase_matrix, specimen_slot, axis=0)
                # Update our refinement tree store to reflect current state
                self.refinement.update_refinement_treestore()
        # Inform any interested party they need to update their representation
        self.needs_reset.emit()

    def del_specimen_slot_by_object(self, specimen):
        """ Deletes a specimen slot using the actual object """
        try:
            self.del_specimen_slot(self.specimens.index(specimen))
        except ValueError:
            logger.exception("Caught a ValueError when deleting a specimen from  mixture '%s'" % self.name)

    # ------------------------------------------------------------
    #      Refinement stuff:
    # ------------------------------------------------------------
    def set_data_object(self, mixture, calculate=False):
        """
            Sets the fractions, scales and bgshifts of this mixture.
        """
        if mixture is not None:
            with self.needs_update.ignore():
                with self.data_changed.hold_and_emit():
                    self.fractions[:] = list(mixture.fractions)
                    self.scales[:] = list(mixture.scales)
                    self.bgshifts[:] = list(mixture.bgshifts)

                    # reset flag when needed:
                    mixture.calculated = mixture.calculated and not calculate
                    mixture = self.optimizer.calculate(mixture)

                    for i, (specimen_data, specimen) in enumerate(zip(mixture.specimens, self.specimens)):
                        if specimen is not None:
                            with specimen.data_changed.ignore():
                                specimen.update_pattern(
                                    specimen_data.total_intensity,
                                    specimen_data.scaled_phase_intensities,
                                    self.phase_matrix[i, :]
                                )

    def optimize(self):
        """
            Optimize the current solution (fractions, scales, bg shifts & calculate
            phase intensities)
        """
        with self.needs_update.ignore():
            with self.data_changed.hold():
                # no need to re-calculate, is already done by the optimization
                self.set_data_object(self.optimizer.optimize())

    def apply_current_data_object(self):
        """
            Recalculates the intensities using the current fractions, scales
            and bg shifts without optimization
        """
        with self.needs_update.ignore():
            with self.data_changed.hold():
                self.set_data_object(self.data_object, calculate=True)

    # @print_timing
    def update(self):
        """
            Optimizes or re-applies the current mixture 'solution'.
            Effectively re-calculates the entire patterns.
        """
        with self.needs_update.ignore():
            with self.data_changed.hold():
                if self.auto_run:
                    self.optimize()
                else:
                    self.apply_current_data_object()

    # ------------------------------------------------------------
    #      Various other things:
    # ------------------------------------------------------------
    def get_composition_matrix(self):
        """
            Returns a matrix containing the oxide composition for each specimen 
            in this mixture. It uses the COMPOSITION_CONV file for this purpose
            to convert element weights into their oxide weight equivalent.
        """

        # create an atom nr -> (atom name, conversion) mapping
        # this is used to transform most of the elements into their oxides
        atom_conv = OrderedDict()
        with open(settings.DATA_REG.get_file_path("COMPOSITION_CONV"), 'r') as f:
            reader = csv.reader(f)
            next(reader) # skip header
            for row in reader:
                nr, name, fact = row
                atom_conv[int(nr)] = (name, float(fact))

        comps = list()
        for i, row in enumerate(self.phase_matrix):
            comp = dict()
            for j, phase in enumerate(row):
                phase_fract = self.fractions[j]
                for k, component in enumerate(phase.components):
                    comp_fract = phase.probabilities.mW[k] * phase_fract
                    for atom in chain(component.layer_atoms,
                            component.interlayer_atoms):
                        nr = atom.atom_type.atom_nr
                        if nr in atom_conv:
                            wt = atom.pn * atom.atom_type.weight * comp_fract * atom_conv[nr][1]
                            comp[nr] = comp.get(nr, 0.0) + wt
            comps.append(comp)

        final_comps = np.zeros(shape=(len(atom_conv) + 1, len(comps) + 1), dtype='a15')
        final_comps[0, 0] = " "*8
        for j, comp in enumerate(comps):
            fact = 100.0 / sum(comp.values())
            for i, (nr, (oxide_name, conv)) in enumerate(atom_conv.items()):
                wt = comp.get(nr, 0.0) * fact
                # set relevant cells:
                if i == 0:
                    final_comps[i, j + 1] = self.specimens[j].name.ljust(15)[:15]
                if j == 0:
                    final_comps[i + 1, j] = ("%s  " % oxide_name).rjust(8)[:8]
                final_comps[i + 1, j + 1] = ("%.1f" % wt).ljust(15)[:15]

        return final_comps

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import os, sys

# Make sure the current path is used for loading PyXRD modules:
mod = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
if not mod in sys.path:
    sys.path.insert(1, mod)

from pyxrd.core import run_main
run_main()



#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import unittest

from test.utils import create_object_attribute_test

from pyxrd.phases.models import Component

__all__ = [
    'TestComponent',
]

class TestComponent(unittest.TestCase):

    component = None

    def setUp(self):
        self.component = Component()

    def tearDown(self):
        del self.component

    def test_not_none(self):
        self.assertIsNotNone(self.component)

    def test_data_object(self):
        self.assertIsNotNone(self.component.data_object)

    test_name = create_object_attribute_test("component", "name", "Test Name")
    test_d001 = create_object_attribute_test("component", "d001", 0.789)
    test_default_c = create_object_attribute_test("component", "default_c", 0.646)
    test_delta_c = create_object_attribute_test("component", "delta_c", 0.002)
    test_inherit_atom_relations = create_object_attribute_test("component", "inherit_atom_relations", True)
    test_inherit_interlayer_atoms = create_object_attribute_test("component", "inherit_interlayer_atoms", True)
    test_inherit_layer_atoms = create_object_attribute_test("component", "inherit_layer_atoms", True)
    test_inherit_delta_c = create_object_attribute_test("component", "inherit_delta_c", True)
    test_inherit_default_c = create_object_attribute_test("component", "inherit_default_c", True)
    test_inherit_ucp_a = create_object_attribute_test("component", "inherit_ucp_a", True)
    test_inherit_ucp_b = create_object_attribute_test("component", "inherit_ucp_b", True)
    test_inherit_d001 = create_object_attribute_test("component", "inherit_d001", True)

    def _setup_inheritance(self):
        self.component2 = Component()
        self.component2.linked_with = self.component
        self.component2.inherit_atom_relations = True # TODO
        self.component2.inherit_interlayer_atoms = True # TODO
        self.component2.inherit_layer_atoms = True # TODO
        self.component2.inherit_delta_c = True
        self.component2.inherit_default_c = True
        self.component2.inherit_ucp_a = True # TODO
        self.component2.inherit_ucp_b = True # TODO
        self.component2.inherit_d001 = True

    def test_inheritance_for_delta_c(self):
        self._setup_inheritance()

        self.component.delta_c = 0.005
        self.assertEqual(self.component2.delta_c, 0.005)

    def test_inheritance_for_default_c(self):
        self._setup_inheritance()

        self.component.default_c = 0.750
        self.assertEqual(self.component2.default_c, 0.750)

    def test_inheritance_for_d001(self):
        self._setup_inheritance()

        self.component.d001 = 0.750
        self.assertEqual(self.component2.d001, 0.750)

    pass # end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

#TODO make this gtk agnostic

from pkg_resources import resource_filename # @UnresolvedImport
import gi
gi.require_version('Gtk', '3.0')  # @UndefinedVariable
from gi.repository import GdkPixbuf  # @UnresolvedImport

def get_icon_list():
    return [
        GdkPixbuf.Pixbuf.new_from_file(resource_filename(__name__, "icons/pyxrd_icon_16x16.png")), #@UndefinedVariable 
        GdkPixbuf.Pixbuf.new_from_file(resource_filename(__name__, "icons/pyxrd_icon_24x24.png")), #@UndefinedVariable
        GdkPixbuf.Pixbuf.new_from_file(resource_filename(__name__, "icons/pyxrd_icon_32x32.png")), #@UndefinedVariable
        GdkPixbuf.Pixbuf.new_from_file(resource_filename(__name__, "icons/pyxrd_icon_48x48.png")), #@UndefinedVariable
        GdkPixbuf.Pixbuf.new_from_file(resource_filename(__name__, "icons/pyxrd_icon_64x64.png")), #@UndefinedVariable
        GdkPixbuf.Pixbuf.new_from_file(resource_filename(__name__, "icons/pyxrd_icon_128x128.png")) #@UndefinedVariable
    ]



# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

from mvc.support.utils import not_none

class ValueWrapper(object):

    wrappers = []

    @staticmethod
    def register_wrapper(cls=None, position=None):
        """
            Decorator that can be applied to wrapper classes (need to 
            define a wrap_value(label, value, model=None) method).
            Can also be called with a position keyword argument to
            force that wrapper to be at a certain position in the list of
            wrappers.
            Otherwise appends the wrapper to the end of wrapper list.
        """
        def inner(cls, position=None):
            position = not_none(position, len(ValueWrapper.wrappers))
            ValueWrapper.wrappers.insert(position, cls.wrap_value)
        if cls == None:
            return inner
        else:
            return inner(cls, position=position)

    @staticmethod
    def wrap_value(label, val, model=None, verbose=False):
        """This is used to wrap a value to be assigned to a
        property. Depending on the type of the value, different values
        are created and returned. For example, for a list, a
        ListWrapper is created to wrap it, and returned for the
        assignment. model is different from None when the value is
        changed (a model exists). Otherwise, during property creation
        model is None"""

        for wrapper in ValueWrapper.wrappers:
            wrapped = wrapper(label, val, model)
            if wrapped is None:
                continue
            else:
                return wrapped

        return val

    pass #end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.


from contextlib import contextmanager

from mvc.models.base import Model

from pyxrd.generic.controllers import ObjectListStoreController

from pyxrd.mixture.models import InSituBehaviour, insitu_behaviours
from pyxrd.mixture.views import EditInSituBehaviourView


from pyxrd.mixture.views.add_insitu_behaviour_view import AddInSituBehaviourView

from .edit_insitu_behaviour_controller import EditInSituBehaviourController
from .add_insitu_behaviour_controller import AddInSituBehaviourController

class InSituBehavioursController(ObjectListStoreController):

    treemodel_property_name = "behaviours"
    treemodel_class_type = InSituBehaviour
    columns = [ ("Mixture name", "c_name") ]
    delete_msg = "Deleting a mixture is irreverisble!\nAre You sure you want to continue?"
    obj_type_map = [
        (cls, EditInSituBehaviourView, EditInSituBehaviourController)
        for name, cls in list(insitu_behaviours.__dict__.items()) if hasattr(cls, 'Meta') and cls.Meta.concrete
    ]

    def get_behaviours_tree_model(self, *args):
        return self.treemodel

    # ------------------------------------------------------------
    #      GTK Signal handlers
    # ------------------------------------------------------------
    def on_load_object_clicked(self, event):
        pass # cannot load behaviours
    
    def on_save_object_clicked(self, event):
        pass # cannot save behaviours

    def get_new_edit_view(self, obj):
        """
            Gets a new 'edit object' view for the given obj, view and parent
            view.
        """
        if obj == None:
            return self.view.none_view
        else:
            for obj_tp, view_tp, ctrl_tp in self.obj_type_map: # @UnusedVariable
                if isinstance(obj, obj_tp):
                    return view_tp(obj.Meta, parent=self.view)
            raise NotImplementedError("Unsupported object type; subclasses of"
                " TreeControllerMixin need to define an obj_type_map attribute!")

    def create_new_object_proxy(self):
        
        def on_accept(behaviour_type):
            if behaviour_type is not None:
                self.add_object(behaviour_type(parent=self.model))

        # TODO re-use this and reset the COMBO etc.
        self.add_model = Model()
        self.add_view = AddInSituBehaviourView(parent=self.view)
        self.add_ctrl = AddInSituBehaviourController(
            model=self.add_model, view=self.add_view, parent=self,
            callback=on_accept
        )

        self.add_view.present()
        return None

    @contextmanager
    def _multi_operation_context(self):
        with self.model.data_changed.hold():
            yield

    pass # end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from .base import PyXRDModel, ChildModel, DataModel
from .lines import PyXRDLine, CalculatedLine, ExperimentalLine

__all__ = [
    "PyXRDModel", "ChildModel", "DataModel",
    "PyXRDLine", "CalculatedLine", "ExperimentalLine",
]



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from pyxrd.generic.utils import not_none
from pyxrd.refinement.async_evaluatable import AsyncEvaluatable

class RefineAsyncHelper(AsyncEvaluatable):
    """
        Helper class which can help classes having a refiner object
        
    """
    def do_async_evaluation(self, iter_func, eval_func=None, data_func=None, result_func=None):
        assert self.refiner is not None, "RefineAsyncHelper can only work when a refiner is set!"
        eval_func = not_none(eval_func, self.refiner.residual_callback)
        data_func = not_none(data_func, self.refiner.get_data_object)
        result_func = not_none(result_func, self.refiner.update)
        return super(RefineAsyncHelper, self).do_async_evaluation(
            iter_func, eval_func, data_func, result_func
        )


#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

B_DO_PROFILING = False
if B_DO_PROFILING:
    import cProfile, pstats

import logging
logger = logging.getLogger(__name__)

import multiprocessing
import os
import codecs
import numpy as np

"""
ADDED 0.01 NOISE
"""

def get_uniform_solutions(refiner, num):
    """
        Returns `num` solutions (uniformly distributed within their ranges) 
        for the selected parameters.
    """
    start_solutions = np.random.random_sample((num, len(refiner.ref_props)))
    ranges = np.asarray(refiner.ranges, dtype=float)
    return ranges[:, 0] + start_solutions * (ranges[:, 1] - ranges[:, 0])

def run(args):
    """
    This is a simple script that will open a PyXRD project file,
    will run a refinement for a certain mixture, and store the results
    in an overview file and the best solution as a new project file.
    The refinement setup is left unchanged, so be sure you have correctly 
    defined parameter ranges and chosen a good refinement strategy (CMA-ES
    is recommended).
    
    To use this script, launch it using PyXRD's core.py launcher script as:
      python core.py -s pyxrd/scripts/refinement_generator.py "$FILENAME###$I###$J"
    in which: 
        - $FILENAME can be replaced with the absolute path to the actual
          project filename
        - $I is the index of the mixture to refine and
        - $J is the 'trial' number, which is added to the record file and to the
          best solution project file.
        - leave the three # (hashes) where they are, they are used as separators 
        
    You can use this script (e.g. on high-performance computing clusters) to
    run several iterations of the same project. Reasaons why you would want 
    to do this are for benchmarking, checking solution reliability, ... 
    Just change the trial number from e.g. 0 to 49 to have 50 iterations.
    """

    ## TODO:
    ##  - use a uniform distribution of starting solutions:
    ##      xs = np.random.uniform(size=50)
    ##      ys = np.random.uniform(size=50)
    ##      zs = np.random.uniform(size=50)

    ##
    ## When the jobs are submitted, load the project and mixture once,
    ## create the # of staring solutions and store them in a file (using np IO)
    ## Then here we can load them and pick the one we need.
    ##

    if args and args.filename != "":
        logging.info("Proccessing args...")
        project_file, k, mixture_index = tuple(args.filename.split("###", 2))
        base_path = os.path.dirname(args.filename)
        start_solutions_fname = os.path.join(
            base_path,
            "start_solutions %s mixture %s" % (os.path.basename(project_file), mixture_index)
        )
        stop_event = multiprocessing.Event()

        logging.info("Loading project file...")
        from pyxrd.file_parsers.json_parser import JSONParser
        project = JSONParser.parse(project_file)
        logging.info(" ".join(["Running Project", os.path.basename(project_file), "Trial", k]))

        for i, mixture in enumerate(project.mixtures):
            if i == int(mixture_index):

                if B_DO_PROFILING:
                    pr = cProfile.Profile()
                    pr.enable()
                try:
                    with mixture.data_changed.hold():

                        mixture.refinement.update_refinement_treestore()
                        refiner = mixture.refinement.get_refiner()

                        if int(k) == 0: #First run, create solutions & store for later use:
                            start_solutions = get_uniform_solutions(refiner, 50)
                            np.savetxt(start_solutions_fname, start_solutions)
                        else:
                            start_solutions = np.loadtxt(start_solutions_fname)

                        refiner.update(start_solutions[k, ...], iteration=-1)
                        mixture.optimizer.optimize()
                        
                        refiner.refine(stop_event)
                except:
                    raise
                finally:
                    if B_DO_PROFILING:
                        pr.disable()
                        with open("pyxrd_stats", "w+") as f:
                            sortby = 'cumulative'
                            ps = pstats.Stats(pr, stream=f).sort_stats(sortby)
                            ps.print_stats()

                recordf = os.path.basename(project_file).replace(".pyxrd", "")
                recordf = base_path + "/" + "record#" + str(k) + " " + recordf + " " + mixture.name
                with codecs.open(recordf, 'w', 'utf-8') as f:

                    f.write("################################################################################\n")
                    f.write(recordf + "\n")
                    f.write("Mixture " + str(i) + " and trial " + str(k) + "\n")
                    f.write("Property name, initial, best, min, max" + "\n")
                    for j, ref_prop in enumerate(refiner.refinable_properties):
                        line = ", ".join([
                            ref_prop.get_descriptor(),
                            str(refiner.history.initial_solution[j]),
                            str(refiner.history.best_solution[j]),
                            str(ref_prop.value_min),
                            str(ref_prop.value_max),
                        ])
                        f.write(line + "\n")
                    f.write("################################################################################\n")

                    def write_records(f, record_header, records):
                        f.write(", ".join(record_header) + "\n")
                        for record in records:
                            f.write(", ".join(["%.7f" % f for f in record]) + "\n")
                        f.write("################################################################################\n")

                    # Apply found solution and save:
                    refiner.apply_best_solution()
                    mixture.optimizer.optimize()

                    project_file_output = base_path + "/" + os.path.basename(project_file).replace(".pyxrd", "") + " - mixture %s - trial %s.pyxrd" % (str(i), str(k))
                    JSONParser.write(project, project_file_output, zipped=True)

        pass # end



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
logger = logging.getLogger(__name__)

import weakref
from warnings import warn

from mvc import Model
from mvc.models.properties import LabeledProperty, SignalProperty
from mvc.support.utils import pop_kwargs, not_none

class PyXRDModel(Model):
    """
        A UUIDModel with some common PyXRD functionality
    """

    class Meta(Model.Meta):
        @classmethod
        def get_refinable_properties(cls):
            if not hasattr(cls, "all_properties"):
                raise RuntimeError("Meta class '%s' has not been initialized" \
                    " properly: 'all_properties' is not set!" % type(self))
            else:
                return [attr for attr in cls.all_properties if getattr(attr, "refinable", False)]

        pass # end of class

    # ------------------------------------------------------------
    #      Methods & functions
    # ------------------------------------------------------------
    def pop_kwargs(self, kwargs, *keys):
        return pop_kwargs(kwargs, *keys)

    def get_kwarg(self, fun_kwargs, default, *keywords):
        """
        Convenience function to get a certain keyword 'kw' value from the passed
        keyword arguments 'fun_kwargs'. If the key 'kw' is not in 'fun_kwargs'
        a list of deprecated keywords to be searched for can be passed as an
        optional argument list 'depr_kws'. If one of these is found, its value
        is returned and a deprecation warning is emitted. 
        If neither the 'kw' nor any of the 'depr_kws' are found the 'default'
        value is returned. 
        """
        if len(keywords) < 1:
            raise AttributeError("get_kwarg() requires at least one keyword (%d given)" % (len(keywords)))

        value = default
        for i, key in enumerate(keywords[::-1]):
            if key in fun_kwargs:
                value = not_none(fun_kwargs[key], default)
                if i != 0:
                    warn("The use of the keyword '%s' is deprecated for %s!" %
                        (key, type(self)), DeprecationWarning)
        return value

    def get_list(self, fun_kwargs, default, *keywords, **kwargs):
        """
            Convenience function to get a 'list' type keyword. Supports
            deprecated serialized ObjectListStores (replaced by regular lists).
        """
        return self.parse_list(self.get_kwarg(fun_kwargs, default, *keywords), **kwargs)

    def parse_list(self, list_arg, **kwargs):
        """
            Parses a list keyword argument (be it an actual list, or a
            former JSON-serialized ObjectListStore object).
        """
        if isinstance(list_arg, dict) and "type" in list_arg:
            list_arg = list_arg["properties"]["model_data"]
        if list_arg is not None:
            return [
                self.parse_init_arg(json_obj, None, child=True, **kwargs)
                for json_obj in list_arg
             ]
        else:
            return list()

    pass # end of class

class ChildModel(PyXRDModel):
    """
        A PyXRDModel with child-parent relation support.
    """

    # MODEL INTEL:
    class Meta(PyXRDModel.Meta):

        @classmethod
        def get_inheritable_properties(cls): # TODO MOVE THIS TO THE CHILD MODEL!!
            if not hasattr(cls, "all_properties"):
                raise RuntimeError("Meta class '%s' has not been initialized" \
                    " properly: 'all_properties' is not set!" % type(cls))
            else:
                return [attr for attr in cls.all_properties if getattr(attr, "inheritable", False)]

    # SIGNALS:
    removed = SignalProperty()
    added = SignalProperty()

    # PROPERTIES:
    __parent = None
    def __get_parent(self):
        if callable(self.__parent):
            return self.__parent()
        else:
            return self.__parent
    def __set_parent(self, value):
        if not self.parent == value:
            if self.parent is not None:
                self.removed.emit()
            try:
                self.__parent = weakref.ref(value, self.__on_parent_finalize)
            except TypeError:
                self.__parent = value
            if self.parent is not None:
                self.added.emit()
    def __on_parent_finalize(self, ref):
        self.removed.emit()
        self.__parent = None
    parent = LabeledProperty(fget=__get_parent, fset=__set_parent)

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, parent=None, *args, **kwargs):
        super(ChildModel, self).__init__(*args, **kwargs)
        self.parent = parent

    pass # end of class

class DataModel(ChildModel):
    """
        A ChildModel with support for having 'calculation data' and 'visual data'            
    """

    # SIGNALS:
    data_changed = SignalProperty()
    visuals_changed = SignalProperty()

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

import sys, html
from contextlib import contextmanager

import gi
from mvc.adapters.gtk_support.widgets.threaded_task_box import ThreadedTaskBox
from mvc.support.gui_loop import run_when_idle, add_timeout_call,\
    remove_timeout_call
from mvc.support.cancellable_thread import CancellableThread
gi.require_version('Gtk', '3.0')  # @UndefinedVariable
from gi.repository import Gtk, GObject  # @UnresolvedImport

from .message_dialog import MessageDialog
from .file_chooser_dialog import FileChooserDialog

class DialogFactory(object):

    # ------------------------------------------------------------
    #      File dialog creators
    # ------------------------------------------------------------

    @staticmethod
    def get_file_dialog_from_context(context, parent=None, **kwargs):
        return FileChooserDialog(context, parent=parent, **kwargs)

    @staticmethod
    def get_file_dialog(
            action, title, parent=None,
            current_name=None, current_folder=None,
            extra_widget=None, filters=[],
            multiple=False, confirm_overwrite=True, persist=False):
        """ Generic file dialog creator """
        return FileChooserDialog(
            title=title, action=action, parent=parent,
            buttons=(Gtk.STOCK_CANCEL, Gtk.ResponseType.REJECT,
                     Gtk.STOCK_OK, Gtk.ResponseType.ACCEPT),
            current_name=current_name, current_folder=current_folder,
            extra_widget=extra_widget, filters=filters,
            multiple=multiple, confirm_overwrite=confirm_overwrite, persist=persist
        )

    @staticmethod
    def get_save_dialog(
            title, parent=None,
            current_name=None, current_folder=None,
            extra_widget=None, filters=[],
            confirm_overwrite=True, persist=False):
        """ Save file dialog creator """
        # Forces save action
        # Does not allow selecting multiple files
        return DialogFactory.get_file_dialog(
            action=Gtk.FileChooserAction.SAVE, title=title, parent=parent,
            current_name=current_name, current_folder=current_folder,
            extra_widget=extra_widget, filters=filters,
            multiple=False, confirm_overwrite=confirm_overwrite, persist=persist)

    @staticmethod
    def get_load_dialog(
            title, parent=None,
            current_name=None, current_folder=None,
            extra_widget=None, filters=[],
            multiple=True, persist=False):
        """ Load file dialog creator """
        # Forces open action
        # Disables overwrite confirmation (doesn't matter really)
        return DialogFactory.get_file_dialog(
            action=Gtk.FileChooserAction.OPEN, title=title, parent=parent,
            current_name=current_name, current_folder=current_folder,
            extra_widget=extra_widget, filters=filters,
            multiple=multiple, confirm_overwrite=False, persist=persist)

    # ------------------------------------------------------------
    #      Message dialog creators
    # ------------------------------------------------------------
    @staticmethod
    def get_message_dialog(message, type, buttons=Gtk.ButtonsType.YES_NO, persist=False, parent=None, title=None):  # @ReservedAssignment
        """ Generic message dialog creator """
        return MessageDialog(
            message=message,
            parent=parent,
            type=type,
            flags=Gtk.DialogFlags.DESTROY_WITH_PARENT,
            buttons=buttons,
            persist=persist,
            title=title)

    @staticmethod
    def get_confirmation_dialog(message, persist=False, parent=None, title=None):
        """ Confirmation dialog creator """
        return DialogFactory.get_message_dialog(
            message,
            parent=parent,
            type=Gtk.MessageType.WARNING,
            persist=persist,
            title=title
        )

    @staticmethod
    def get_information_dialog(message, persist=False, parent=None, title=None):
        """ Information dialog creator """
        return DialogFactory.get_message_dialog(
            message,
            parent=parent,
            type=Gtk.MessageType.INFO,
            buttons=Gtk.ButtonsType.OK,
            persist=persist,
            title=title
        )

    @staticmethod
    def get_error_dialog(message, persist=False, parent=None, title=None):
        """ Error dialog creator """
        return DialogFactory.get_message_dialog(
            message,
            parent=parent,
            type=Gtk.MessageType.ERROR,
            buttons=Gtk.ButtonsType.OK,
            persist=persist,
            title=title
        )

    # ------------------------------------------------------------
    #      Custom dialog creator
    # ------------------------------------------------------------
    @staticmethod
    def get_custom_dialog(content, parent=None):
        window = Gtk.Window()
        window.set_border_width(10)
        window.set_modal(True)
        window.set_transient_for(parent)
        window.connect('delete-event', lambda widget, event: True)
        window.add(content)
        return window

    pass #end of class


    @staticmethod
    @contextmanager  
    def error_dialog_handler(message, parent=None, title=None, reraise=True, print_tb=True):
        """ Context manager that can be used to wrap error-prone code. If an error
        is risen, a dialog will inform the user, optionally the error can be re-raised """
        try:
            yield
        except:
            msg = message.format(html.escape("%s" % sys.exc_info()[1]))
            DialogFactory.get_error_dialog(
               msg, title=title, parent=parent
            ).run()
            if reraise: raise # This should be handled by the default UI bug dialog
            elif print_tb:
                from traceback import print_exc
                print_exc()
                
    @staticmethod
    def get_progress_dialog(action, complete_callback=None, gui_message="Processing ...", toplevel=None):
        """
            Returns a callable that will show a progress dialog
            for the given action - which will be run in a different 
            thread from the GUI.
            The action is expected to take a single argument: `status_dict`
            which is used to format the gui_message (new-style formatting).
            toplevel is the top level window.
            complete_callback is called when the action has completed with
            its return value.
            When interrupted or cancelled by the user, the dialog just hides.  
        """
        def run_action_and_show_progress():

            taskgui = ThreadedTaskBox()
            window = DialogFactory.get_custom_dialog(
                 taskgui, parent=toplevel)

            # Status:
            status_dict = dict()

            # Task:
            def load_peak_thresholds(stop=None):
                action(status_dict)

            # Cancel & stop events:
            def on_interrupted(*args, **kwargs):
                window.hide()

            # Status label update:
            def gui_callback():
                taskgui.set_status(gui_message.format(**status_dict))
                return True
            add_timeout_call(250, gui_callback)

            # Complete event:
            @run_when_idle
            def on_complete(*args, **kwargs):
                remove_timeout_call(gui_callback)
                taskgui.stop()
                window.destroy()
                if callable(complete_callback):
                    complete_callback(*args, **kwargs)
                
            # Run thread:
            thread = CancellableThread(load_peak_thresholds, on_complete)
            thread.start()
            
            # Run task box:
            taskgui.connect("cancelrequested", on_interrupted)
            taskgui.connect("stoprequested", on_interrupted)
            taskgui.set_status("Loading ...")
            taskgui.start()
            window.show_all()

        return run_action_and_show_progress



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
logger = logging.getLogger(__name__)

import numpy as np
from scipy.integrate import trapz
from scipy.interpolate import UnivariateSpline

from mvc.models.properties.tools import modify
from mvc.models.properties import (
    FloatProperty, LabeledProperty, IntegerProperty, FloatChoiceProperty,
    IntegerChoiceProperty, SetActionMixin, SignalMixin
)

from pyxrd.data import settings
from pyxrd.generic.io import storables
from pyxrd.calculations.math_tools import smooth, add_noise

from pyxrd.generic.models.base import DataModel

from .pyxrd_line import PyXRDLine

@storables.register()
class ExperimentalLine(PyXRDLine):

    # MODEL INTEL:
    class Meta(PyXRDLine.Meta):
        store_id = "ExperimentalLine"

    specimen = property(DataModel.parent.fget, DataModel.parent.fset)

    # PROPERTIES:

    #: The line color
    color = modify(PyXRDLine.color,
        default=settings.EXPERIMENTAL_COLOR,
        inherit_from="parent.parent.display_exp_color"
    )
    #: The linewidth in points
    lw = modify(PyXRDLine.lw,
        default=settings.EXPERIMENTAL_LINEWIDTH,
        inherit_from="parent.parent.display_exp_lw"
    )

    #: A short string describing the (matplotlib) linestyle
    ls = modify(PyXRDLine.ls,
        default=settings.EXPERIMENTAL_LINESTYLE,
        inherit_from="parent.parent.display_exp_ls"
    )

    #: A short string describing the (matplotlib) marker
    marker = modify(PyXRDLine.marker,
        default=settings.EXPERIMENTAL_MARKER,
        inherit_from="parent.parent.display_exp_marker"
    )

    #: The value to cap the pattern at (in raw values)
    cap_value = FloatProperty(
        default=0.0, text="Cap value",
        persistent=True, visible=True, widget_type="float_entry",
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )

    @property
    def max_display_y(self):
        max_value = super(ExperimentalLine, self).max_display_y
        # Only cap single and multi-line patterns, not 2D images:
        if self.cap_value > 0 and not (self.num_columns > 2 and len(self.z_data)):
            max_value = min(max_value, self.cap_value)
        return max_value

    ###########################################################################

    #: The background offset value
    bg_position = FloatProperty(
        default=0.0, text="Background offset",
        persistent=False, visible=True, widget_type="float_entry",
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )

    #: The background scale
    bg_scale = FloatProperty(
        default=1.0, text="Background scale",
        persistent=False, visible=True, widget_type="float_entry",
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )

    #: The background pattern or None for linear patterns
    bg_pattern = LabeledProperty(
        default=None, text="Background pattern",
        persistent=False, visible=False,
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )

    #: The background type: pattern or linear
    bg_type = IntegerChoiceProperty(
        default=0, text="Background type", choices=settings.PATTERN_BG_TYPES,
        persistent=False, visible=True,
        signal_name="visuals_changed", set_action_name="find_bg_position",
        mix_with=(SignalMixin, SetActionMixin,)
    )

    def get_bg_type_lbl(self):
        return settings.PATTERN_BG_TYPES[self.bg_type]

    ###########################################################################

    #: Pattern smoothing type
    smooth_type = IntegerChoiceProperty(
        default=0, text="Smooth type", choices=settings.PATTERN_SMOOTH_TYPES,
        persistent=False, visible=True,
        signal_name="visuals_changed", set_action_name="setup_smooth_variables",
        mix_with=(SignalMixin, SetActionMixin,)
    )

    smooth_pattern = None

    #: The smooth degree
    smooth_degree = IntegerProperty(
        default=0, text="Smooth degree",
        persistent=False, visible=True,
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )

    ###########################################################################

    #: The noise fraction to add
    noise_fraction = FloatProperty(
        default=0.0, text="Noise fraction",
        persistent=False, visible=True, widget_type="spin",
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )

    ###########################################################################

    #: The pattern shift correction value
    shift_value = FloatProperty(
        default=0.0, text="Shift value",
        persistent=False, visible=True, widget_type="float_entry",
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )

    #: Shift reference position
    shift_position = FloatChoiceProperty(
        default=0.42574, text="Shift position", choices=settings.PATTERN_SHIFT_POSITIONS,
        persistent=False, visible=True,
        signal_name="visuals_changed", set_action_name="setup_shift_variables",
        mix_with=(SignalMixin, SetActionMixin,)
    )

    ###########################################################################

    #: The peak properties calculation start position
    peak_startx = FloatProperty(
        default=0.0, text="Peak properties start position",
        persistent=False, visible=True, widget_type="float_entry",
        set_action_name="update_peak_properties",
        mix_with=(SetActionMixin,)
    )

    #: The peak properties calculation end position
    peak_endx = FloatProperty(
        default=0.0, text="Peak properties end position",
        persistent=False, visible=True, widget_type="float_entry",
        set_action_name="update_peak_properties",
        mix_with=(SetActionMixin,)
    )

    #: The peak fwhm value
    peak_fwhm_result = FloatProperty(
        default=0.0, text="Peak FWHM value",
        persistent=False, visible=True, widget_type="label",
    )
    
    #: The peak area value
    peak_area_result = FloatProperty(
        default=0.0, text="Peak area value",
        persistent=False, visible=True, widget_type="label",
    )

    #: The patterns peak properties are calculated from
    peak_properties_pattern = LabeledProperty(
        default=None, text="Peak properties pattern",
        persistent=False, visible=False,
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )

    ###########################################################################

    #: The strip peak start position
    strip_startx = FloatProperty(
        default=0.0, text="Strip peak start position",
        persistent=False, visible=True, widget_type="float_entry",
        set_action_name="update_strip_pattern",
        mix_with=(SetActionMixin,)
    )

    #: The strip peak end position
    strip_endx = FloatProperty(
        default=0.0, text="Strip peak end position",
        persistent=False, visible=True, widget_type="float_entry",
        set_action_name="update_strip_pattern",
        mix_with=(SetActionMixin,)
    )

    #: The stripped peak pattern
    stripped_pattern = LabeledProperty(
        default=None, text="Strip peak pattern",
        persistent=False, visible=False,
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )

    #: The stripped peak pattern noise
    noise_level = FloatProperty(
        default=0.0, text="Strip peak noise level",
        persistent=False, visible=True, widget_type="float_entry",
        set_action_name="update_strip_pattern_noise",
        mix_with=(SetActionMixin,)
    )

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, cap_value=0.0, *args, **kwargs):
        """
            Valid keyword arguments for a ExperimentalLine are:
                cap_value: the value (in raw counts) at which to cap
                 the experimental pattern  
        """
        super(ExperimentalLine, self).__init__(*args, **kwargs)
        self.cap_value = cap_value

    # ------------------------------------------------------------
    #      Background Removal
    # ------------------------------------------------------------
    def remove_background(self):
        with self.data_changed.hold_and_emit():
            bg = None
            if self.bg_type == 0:
                bg = self.bg_position
            elif self.bg_type == 1 and self.bg_pattern is not None and not (self.bg_position == 0 and self.bg_scale == 0):
                bg = self.bg_pattern * self.bg_scale + self.bg_position
            if bg is not None and self.data_y.size > 0:
                self.data_y[:, 0] -= bg
            self.clear_bg_variables()

    def find_bg_position(self):
        try:
            self.bg_position = np.min(self.data_y)
        except ValueError:
            return 0.0

    def clear_bg_variables(self):
        with self.visuals_changed.hold_and_emit():
            self.bg_pattern = None
            self.bg_scale = 0.0
            self.bg_position = 0.0

    # ------------------------------------------------------------
    #       Data Smoothing
    # ------------------------------------------------------------
    def smooth_data(self):
        with self.data_changed.hold_and_emit():
            if self.smooth_degree > 0:
                degree = int(self.smooth_degree)
                self.data_y[:, 0] = smooth(self.data_y[:, 0], degree)
            self.smooth_degree = 0.0

    def setup_smooth_variables(self):
        with self.visuals_changed.hold_and_emit():
            self.smooth_degree = 5.0

    def clear_smooth_variables(self):
        with self.visuals_changed.hold_and_emit():
            self.smooth_degree = 0.0

    # ------------------------------------------------------------
    #       Noise adding
    # ------------------------------------------------------------
    def add_noise(self):
        with self.data_changed.hold_and_emit():
            if self.noise_fraction > 0:
                noisified = add_noise(self.data_y[:, 0], self.noise_fraction)
                self.set_data(self.data_x, noisified)
            self.noise_fraction = 0.0

    def clear_noise_variables(self):
        with self.visuals_changed.hold_and_emit():
            self.noise_fraction = 0.0

    # ------------------------------------------------------------
    #       Data Shifting
    # ------------------------------------------------------------
    def shift_data(self):
        with self.data_changed.hold_and_emit():
            if self.shift_value != 0.0:
                if settings.PATTERN_SHIFT_TYPE == "Linear":
                    self.data_x = self.data_x - self.shift_value
                    if self.specimen is not None:
                        with self.specimen.visuals_changed.hold():
                            for marker in self.specimen.markers:
                                marker.position = marker.position - self.shift_value
                elif settings.PATTERN_SHIFT_TYPE == "Displacement":
                    position = self.specimen.goniometer.get_t_from_nm(self.shift_position)
                    displacement = 0.5 * self.specimen.goniometer.radius * self.shift_value / np.cos(position / 180 * np.pi)
                    correction = 2 * displacement * np.cos(self.data_x / 2 / 180 * np.pi) / self.specimen.goniometer.radius
                    self.data_x = self.data_x - correction

            self.shift_value = 0.0

    def setup_shift_variables(self):
        with self.visuals_changed.hold_and_emit():
            position = self.specimen.goniometer.get_2t_from_nm(self.shift_position)
            if position > 0.1:
                max_x = position + 0.5
                min_x = position - 0.5
                condition = (self.data_x >= min_x) & (self.data_x <= max_x)
                section_x, section_y = np.extract(condition, self.data_x), np.extract(condition, self.data_y[:, 0])
                try:
                    #TODO to exclude noise it'd be better to first interpolate
                    # or smooth the data and then find the max.
                    actual_position = section_x[np.argmax(section_y)]
                except ValueError:
                    actual_position = position
                self.shift_value = actual_position - position

    def clear_shift_variables(self):
        with self.visuals_changed.hold_and_emit():
            self.shift_value = 0

    # ------------------------------------------------------------
    #       Peak area calculation
    # ------------------------------------------------------------
    peak_bg_slope = 0.0
    avg_starty = 0.0
    avg_endy = 0.0
    def update_peak_properties(self):
        with self.visuals_changed.hold_and_emit():
            if self.peak_endx < self.peak_startx:
                self.peak_endx = self.peak_startx + 1.0
                return # previous line will have re-invoked this method

            # calculate average starting point y value
            condition = (self.data_x >= self.peak_startx - 0.1) & (self.data_x <= self.peak_startx + 0.1)
            section = np.extract(condition, self.data_y[:, 0])
            self.avg_starty = np.min(section)

            # calculate average ending point y value
            condition = (self.data_x >= self.peak_endx - 0.1) & (self.data_x <= self.peak_endx + 0.1)
            section = np.extract(condition, self.data_y[:, 0])
            self.avg_endy = np.min(section)

            # Calculate new bg slope
            self.peak_bg_slope = (self.avg_starty - self.avg_endy) / (self.peak_startx - self.peak_endx)

            # Get the x-values in between start and end point:
            condition = (self.data_x >= self.peak_startx) & (self.data_x <= self.peak_endx)
            section_x = np.extract(condition, self.data_x)
            section_y = np.extract(condition, self.data_y)
            bg_curve = (self.peak_bg_slope * (section_x - self.peak_startx) + self.avg_starty)

            #Calculate the peak area:
            self.peak_area_result = abs(trapz(section_y, x=section_x) - trapz(bg_curve, x=section_x))

            # create a spline of of the peak (shifted down by half of its maximum)
            fwhm_curve = section_y - bg_curve
            peak_half_max = np.max(fwhm_curve)*0.5
            spline = UnivariateSpline(section_x, fwhm_curve-peak_half_max, s=0)
            roots = spline.roots() # find the roots = where the splin = 0
            self.peak_fwhm_result = np.abs(roots[0] - roots[-1]) if (len(roots) >= 2) else 0

            # Calculate the new y-values: x values, bg_curve y values, original pattern y values, x values for the FWHM, y values for the FWHM
            self.peak_properties_pattern = (section_x, bg_curve, section_y, roots, spline(roots)+peak_half_max)

    def clear_peak_properties_variables(self):
        with self.visuals_changed.hold_and_emit():
            self._peak_startx = 0.0
            self._peak_properties_pattern = None
            self._peak_endx = 0.0
            self.peak_properties = 0.0

    # ------------------------------------------------------------
    #       Peak stripping
    # ------------------------------------------------------------
    def strip_peak(self):
        with self.data_changed.hold_and_emit():
            if self.stripped_pattern is not None:
                stripx, stripy = self.stripped_pattern
                indeces = ((self.data_x >= self.strip_startx) & (self.data_x <= self.strip_endx)).nonzero()[0]
                np.put(self.data_y[:, 0], indeces, stripy)
            self._strip_startx = 0.0
            self._stripped_pattern = None
            self.strip_endx = 0.0

    strip_slope = 0.0
    avg_starty = 0.0
    avg_endy = 0.0
    block_strip = False

    def update_strip_pattern_noise(self):
        with self.visuals_changed.hold_and_emit():
            # Get the x-values in between start and end point:
            condition = (self.data_x >= self.strip_startx) & (self.data_x <= self.strip_endx)
            section_x = np.extract(condition, self.data_x)

            # Calculate the new y-values, add noise according to noise_level
            noise = self.avg_endy * 2 * (np.random.rand(*section_x.shape) - 0.5) * self.noise_level
            section_y = (self.strip_slope * (section_x - self.strip_startx) + self.avg_starty) + noise
            self.stripped_pattern = (section_x, section_y)

    def update_strip_pattern(self):
        with self.visuals_changed.hold_and_emit():
            if self.strip_endx < self.strip_startx:
                self.strip_endx = self.strip_startx + 1.0
                return # previous line will have re-invoked this method

            if not self.block_strip:
                self.block_strip = True

                # calculate average starting point y value
                condition = (self.data_x >= self.strip_startx - 0.1) & (self.data_x <= self.strip_startx + 0.1)
                section = np.extract(condition, self.data_y[:, 0])
                self.avg_starty = np.average(section)
                noise_starty = 2 * np.std(section) / self.avg_starty

                # calculate average ending point y value
                condition = (self.data_x >= self.strip_endx - 0.1) & (self.data_x <= self.strip_endx + 0.1)
                section = np.extract(condition, self.data_y[:, 0])
                self.avg_endy = np.average(section)
                noise_endy = 2 * np.std(section) / self.avg_starty

                # Calculate new slope and noise level
                self.strip_slope = (self.avg_starty - self.avg_endy) / (self.strip_startx - self.strip_endx)
                self.noise_level = (noise_starty + noise_endy) * 0.5

                self.update_strip_pattern_noise()

    def clear_strip_variables(self):
        with self.visuals_changed.hold_and_emit():
            self._strip_startx = 0.0
            self._strip_pattern = None
            self.strip_start_x = 0.0

    pass # end of class



#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from .base import AbstractTestProbModel

from test.utils import create_object_attribute_test

from pyxrd.probabilities.models.R1models import *
import unittest

__all__ = [
    'TestR1G2Model',
    'TestR1G3Model',
    'TestR1G4Model',
]

class TestR1G2Model(AbstractTestProbModel, unittest.TestCase):

    prob_model_type = R1G2Model

    test_W1 = create_object_attribute_test("prob_model", "W1", 0.7)
    test_P11_or_P22 = create_object_attribute_test("prob_model", "P11_or_P22", 0.7)

    pass # end of class

class TestR1G3Model(AbstractTestProbModel, unittest.TestCase):

    prob_model_type = R1G3Model

    test_W1 = create_object_attribute_test("prob_model", "W1", 0.7)
    test_P11_or_P22 = create_object_attribute_test("prob_model", "P11_or_P22", 0.7)
    test_G1 = create_object_attribute_test("prob_model", "G1", 0.7)
    test_G2 = create_object_attribute_test("prob_model", "G2", 0.7)
    test_G3 = create_object_attribute_test("prob_model", "G3", 0.7)
    test_G4 = create_object_attribute_test("prob_model", "G4", 0.7)

    pass # end of class

class TestR1G4Model(AbstractTestProbModel, unittest.TestCase):

    prob_model_type = R1G4Model

    test_W1 = create_object_attribute_test("prob_model", "W1", 0.7)
    test_P11_or_P22 = create_object_attribute_test("prob_model", "P11_or_P22", 0.7)
    test_R1 = create_object_attribute_test("prob_model", "R1", 0.7)
    test_R2 = create_object_attribute_test("prob_model", "R2", 0.7)
    test_G1 = create_object_attribute_test("prob_model", "G1", 0.7)
    test_G2 = create_object_attribute_test("prob_model", "G2", 0.7)
    test_G11 = create_object_attribute_test("prob_model", "G11", 0.7)
    test_G12 = create_object_attribute_test("prob_model", "G12", 0.7)
    test_G21 = create_object_attribute_test("prob_model", "G21", 0.7)
    test_G22 = create_object_attribute_test("prob_model", "G22", 0.7)
    test_G31 = create_object_attribute_test("prob_model", "G31", 0.7)
    test_G22 = create_object_attribute_test("prob_model", "G32", 0.7)

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from .namespace import project_parsers

from .json_project_parser import JSONProjectParser
from .sybilla_project_parser import SybillaProjectParser


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import numpy as np

from .base_parser import BaseParser

class ASCIIParser(BaseParser):
    """
        ASCII Parser
    """
    description = "ASCII data"
    mimetypes = ["text/plain", ]
    can_write = True

    @classmethod
    def get_last_line(cls, f):
        i = -1
        f.seek(0)
        for i, l in enumerate(f):
            pass
        return i + 1, l

    @classmethod
    def write(cls, filename, x, ys, header="", delimiter=",", **kwargs):
        """
            Writes the header to the first line, and will write x, y1, ..., yn
            rows for each column inside the x and ys arguments.
            Header argument should not include a newline, and can be a string or
            any iterable containing strings.
        """
        f = open(filename, 'w')
        if not isinstance(header, str):
            header = delimiter.join(header) # assume this is an iterable
        f.write("%s\n" % header)
        np.savetxt(f, np.insert(ys, 0, x, axis=0).transpose(), fmt="%.8f", delimiter=delimiter)
        f.close()

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from pkg_resources import resource_filename  # @UnresolvedImport

from math import isnan

from mvc.adapters.gtk_support.widgets.threaded_task_box import ThreadedTaskBox

from pyxrd.generic.views import  DialogView
from pyxrd.generic.utils import not_none

class RefinementView(DialogView):
    title = "Refine Phase Parameters"
    subview_builder = resource_filename(__name__, "glade/refinement.glade")
    subview_toplevel = "refine_params"
    modal = True

    refine_status_builder = resource_filename(__name__, "glade/refine_status.glade")
    refine_status_toplevel = "tbl_refine_info"
    refine_status_container = "refine_status_box"

    refine_spin_container = "refine_spin_box"
    refine_spin_box = None

    refine_method_builder = resource_filename(__name__, "glade/refine_method.glade")
    refine_method_toplevel = "tbl_refine_method"
    refine_method_container = "refine_method_box"

    def __init__(self, *args, **kwargs):
        super(RefinementView, self).__init__(*args, **kwargs)

        # Add the status box
        self._builder.add_from_file(self.refine_status_builder)
        self._add_child_view(self[self.refine_status_toplevel], self[self.refine_status_container])

        # Add the method and options box
        self._builder.add_from_file(self.refine_method_builder)
        self._add_child_view(self[self.refine_method_toplevel], self[self.refine_method_container])

        # Add the refinement thread box
        self.refine_spin_box = ThreadedTaskBox()
        self._add_child_view(self.refine_spin_box, self[self.refine_spin_container])
        self.hide_refinement_info()

    def connect_cancel_request(self, callback):
        return self.refine_spin_box.connect("cancelrequested", callback)

    def show_refinement_info(self,):
        self["hbox_actions"].set_sensitive(False)
        self["btn_auto_restrict"].set_sensitive(False)
        self[self.refine_method_toplevel].set_sensitive(False)
        self["refinables"].set_visible(False)
        self["refinables"].set_no_show_all(True)

        self[self.refine_status_toplevel].show_all()

    def hide_refinement_info(self):
        self[self.refine_status_toplevel].hide()

        self["hbox_actions"].set_sensitive(True)
        self["btn_auto_restrict"].set_sensitive(True)
        self[self.refine_method_toplevel].set_sensitive(True)
        self["refinables"].set_visible(True)
        self["refinables"].set_no_show_all(False)

    def update_refinement_info(self, current_rp=None, message=None, server_status=None):
        if not isnan(current_rp):
            self["current_residual"].set_text("%.2f" % current_rp)
        self["message"].set_text(not_none(message, ""))
        self.update_server_status(server_status)
        
    def update_server_status(self, server_status):
        color, title, descr = server_status
        self["lbl_server_status"].set_markup("<span foreground=\"%s\">%s</span>" % (color, title))
        self["lbl_server_status"].set_property("tooltip-text", descr)
        self["lbl_server_status"].set_property("tooltip-text", descr)
        self["lbl_server_status"].set_tooltip_text(descr)

    def update_refinement_status(self, status):
        self.refine_spin_box.set_status(status)

    def start_spinner(self):
        self.refine_spin_box.start()

    def stop_spinner(self):
        self.refine_spin_box.stop()

    pass # end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import os
from datetime import date

import numpy as np

from pyxrd.generic.io.utils import get_case_insensitive_glob
from pyxrd.generic.utils import u

from ..base_parser import BaseParser
from .namespace import xrd_parsers
from .xrd_parser_mixin import XRDParserMixin

@xrd_parsers.register_parser()
class CPIParser(XRDParserMixin, BaseParser):
    """
        ASCII Sietronics *.CPI format parser
    """

    description = "Sietronics *.CPI"
    extensions = get_case_insensitive_glob("*.CPI", "*.CPD", "*.CPS")

    @classmethod
    def _parse_header(cls, filename, fp, data_objects=None, close=False):
        f = fp
        
        try:
            basename = u(os.path.basename(filename))
        except:
            basename = None
        
        # Adapt XRDFile list
        data_objects = cls._adapt_data_object_list(data_objects, num_samples=1)

        # Move to the start of the file
        f.seek(0)
        # Skip a line: file type header
        f.readline()
        # Read data limits
        twotheta_min = float(f.readline().replace(",", ".").strip())
        twotheta_max = float(f.readline().replace(",", ".").strip())
        twotheta_step = float(f.readline().replace(",", ".").strip())
        twotheta_count = int((twotheta_max - twotheta_min) / twotheta_step)
        # Read target element name
        target_type = f.readline()
        # Read wavelength
        alpha1 = float(f.readline().replace(",", ".").strip())
        # Read up to SCANDATA and keep track of the line before,
        # it contains the sample description
        name = ""
        while True:
            line = f.readline().strip()
            if line == "SCANDATA" or line == "":
                data_start = f.tell()
                break;
            else:
                name = line

        data_objects[0].update(
            filename=basename,
            name=name,
            target_type=target_type,
            alpha1=alpha1,
            twotheta_min=twotheta_min,
            twotheta_max=twotheta_max,
            twotheta_step=twotheta_step,
            twotheta_count=twotheta_count,
            data_start=data_start,
        )

        if close: f.close()
        return data_objects

    @classmethod
    def _parse_data(cls, filename, fp, data_objects=None, close=False):
        f = fp

        # CPI files are singletons, so no need to iterate over the list,
        # there is only one data object instance:
        if data_objects[0].data == None:
            data_objects[0].data = []

        if f is not None:
            f.seek(data_objects[0].data_start)
            n = 0
            while n <= data_objects[0].twotheta_count:
                line = f.readline().strip("\n").replace(",", ".")
                if line != "":
                    data_objects[0].data.append([ float(data_objects[0].twotheta_min + data_objects[0].twotheta_step * n), float(line) ])
                n += 1

        data_objects[0].data = np.array(data_objects[0].data)

        if close: f.close()
        return data_objects

    @classmethod
    def write(cls, filename, x, ys, radiation="Cu", wavelength=1.54060, tps=48.0, sample="", **kwargs):
        """
            Writes a SIETRONICS cpi text file. x and ys should be numpy arrays.
        """
        start_angle = x[0]
        end_angle = x[-1]
        step_size = (end_angle - start_angle) / (x.size - 1)
        with open(filename, 'w') as f:
            f.write("SIETRONICS XRD SCAN\n")
            f.write("%.4f\n" % start_angle)
            f.write("%.4f\n" % end_angle)
            f.write("%.4f\n" % step_size)
            f.write("%s\n" % radiation)
            f.write("%.5f\n" % wavelength)
            f.write("%s\n" % date.today().strftime('%d/%m/%y %H:%M:%S'))
            f.write("%.1f\n" % tps)
            f.write("%s\n" % sample)
            f.write("SCANDATA\n")
            for y in ys[0, :]:
                f.write("%.7f\n" % y)

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from .base_group_parser import BaseGroupBarser

class ParserNamespace(object):
    """
        ParserNamespace instances can be used to register parsers and bundle
        them for easier retrieval of file filters etc.
    """

    def __init__(self, name, group_description="All supported files"):
        self._parsers = []
        self.name = name
        self.group_description = group_description

    def register_parser(self, first=False):
        """ 
            Register a parsers to this namespace
        """
        def wrapped_register(cls):
            if first:
                self._parsers.insert(0, cls)
            else:
                self._parsers.append(cls)
            self._update_group_parser()
            return cls
        return wrapped_register

    def get_file_filters(self):
        """ 
            Returns all the file filter object for the parsers registered in 
            this namespace 
        """
        for parser in [self._group_parser, ] + self._parsers:
            yield parser.file_filter

    def get_export_file_filters(self):
        for parser in self._parsers:
            if parser.can_write:
                yield parser.file_filter

    def get_import_file_filters(self):
        for parser in [self._group_parser, ] + self._parsers:
            if parser.can_read:
                yield parser.file_filter

    def _update_group_parser(self):
        """
            Factory function for creating BaseGroupParser sub-classes,
            using the namespace's name as the class name and the list of parser
            classes as arguments.
        """
        self._group_parser = type(self.name, (BaseGroupBarser,), dict(
             description=self.group_description,
             parsers=self._parsers
        ))

    pass #end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.


from mvc.models.properties import FloatProperty, BoolProperty

from pyxrd.generic.mathtext_support import mt_range
from pyxrd.generic.io import storables
from pyxrd.generic.utils import not_none
from pyxrd.generic.models.properties import InheritableMixin

from pyxrd.refinement.refinables.properties import RefinableMixin

from .base_models import _AbstractProbability
from mvc.models.properties.action_mixins import SetActionMixin

__all__ = [
    "R1G2Model",
    "R1G3Model",
    "R1G4Model",
]

@storables.register()
class R1G2Model(_AbstractProbability):
    r"""
    Probability model for Reichweite 1 with 2 components.
    
    The 2(=g*(g-1)) independent variables are:
    
    .. math::
        :nowrap:
        
        \begin{flalign*}
            & W_1 \\
            & \text{$P_{11} (W_1 < 0.5)$ or $P_{22} (W_1 > 0.5)$}
        \end{flalign*}
    
    Calculation of the other variables happens as follows:
    
    .. math::
        :nowrap:
        
        \begin{align*}
            & W_2 = 1 – W_1 \\
            & \begin{aligned}
                & \text{$P_{11}$ is given:}  \\
                & \quad P_{12} = 1 - P_{11} \\
                & \quad P_{21} = \frac{W_1 \cdot P_{12}}{W2} \\
                & \quad P_{22} = 1 - P_{21} \\
            \end{aligned}
            \quad \quad
            \begin{aligned}
                & \text{$P_{22}$ is given:} \\
                & \quad P_{21} = 1 - P_{22} \\
                & \quad P_{12} = \frac{W_2 \cdot P_{21}}{W1} \\
                & \quad P_{11} = 1 - P_{12} \\            
            \end{aligned} \\
        \end{align*}
        
    """

    # MODEL METADATA:
    class Meta(_AbstractProbability.Meta):
        store_id = "R1G2Model"

    # PROPERTIES:
    _G = 2

    inherit_W1 = BoolProperty(
        default=False, text="Inherit flag for W1",
        persistent=True, visible=True,
        set_action_name="update",
        mix_with=(SetActionMixin,)
    )

    W1 = FloatProperty(
        default=0.0, text="W1", math_text=r"$W_1$",
        persistent=True, visible=True, refinable=True, store_private=True,
        minimum=0.0, maximum=1.0, is_independent=True, inheritable=True,
        inherit_flag="inherit_W1", inherit_from="parent.based_on.probabilities.W1",
        set_action_name="update",
        mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
    )

    inherit_P11_or_P22 = BoolProperty(
        default=False, text="Inherit flag for P11_or_P22",
        persistent=True, visible=True,
        set_action_name="update",
        mix_with=(SetActionMixin,)
    )

    P11_or_P22 = FloatProperty(
        default=0.0, text="P11_or_P22",
        math_text=r"$P_{11} %s$ or $\newline P_{22} %s$" % (
            mt_range(0.0, "W_1", 0.5),
            mt_range(0.5, "W_1", 1.0)),
        persistent=True, visible=True, refinable=True, store_private=True,
        minimum=0.0, maximum=1.0, is_independent=True, inheritable=True,
        inherit_flag="inherit_P11_or_P22", inherit_from="parent.based_on.probabilities.P11_or_P22",
        set_action_name="update",
        mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
    )

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, W1=0.75, P11_or_P22=0.5, inherit_W1=False, inherit_P11_or_P22=False, *args, **kwargs):
        super(R1G2Model, self).__init__(R=1, *args, **kwargs)

        with self.data_changed.hold():
            self.W1 = not_none(W1, 0.75)
            self.inherit_W1 = inherit_W1
            self.P11_or_P22 = not_none(P11_or_P22, 0.5)
            self.inherit_P11_or_P22 = inherit_P11_or_P22

            self.update()

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def update(self):
        with self.monitor_changes():
            self.mW[0] = self.W1
            self.mW[1] = 1.0 - self.mW[0]
            if self.mW[0] <= 0.5:
                self.mP[0, 0] = self.P11_or_P22
                self.mP[0, 1] = 1.0 - self.mP[0, 0]
                self.mP[1, 0] = self.mW[0] * self.mP[0, 1] / self.mW[1]
                self.mP[1, 1] = 1.0 - self.mP[1, 0]
            else:
                self.mP[1, 1] = self.P11_or_P22
                self.mP[1, 0] = 1.0 - self.mP[1, 1]
                self.mP[0, 1] = self.mW[1] * self.mP[1, 0] / self.mW[0]
                self.mP[0, 0] = 1.0 - self.mP[0, 1]

            self.solve()
            self.validate()

    pass # end of class

@storables.register()
class R1G3Model(_AbstractProbability):
    r"""
    Probability model for Reichweite 1 with 3 components.
    
    The 6 (=g*(g-1)) independent variables are:
    
    .. math::
        :nowrap:
        
        \begin{align*}
                & W_1
                & \text{$P_{11} (W_1 < 0.5)$ or $P_{xx} (W_1 > 0.5)$}
                with P_{xx} = \frac {W_{22} + W_{23} + W_{32} + W_{33} + W_{42} + W_{43}}{W_2 + W_3} \\
                & G_1 = \frac{W_2}{W_2 + W_3}
                & G_2 = \frac{W_{22} + W_{23}}{W_{22} + W_{23} + W_{32} + W_{33}} \\
                & G_3 = \frac{W_{22}}{W_{22} + W_{23}}
                & G_4 = \frac{W_{32}}{W_{32} + W_{33}} \\
        \end{align*}
            
    Calculation of the other variables happens as follows:
    
    
    
    .. math::
        :nowrap:
        
        \begin{align*}
            & \text{Calculate the 'inverted' ratios of $G_2$, $G_3$ and $G_4$ as follows:} \\
            & \quad G_i^{\text{-1}} =
            \begin{cases}
                G_i^{-1} - 1.0, & \text{if } G_i > 0 \\
                0,              & \text{otherwise}
            \end{cases} \quad \forall i \in \left\{ {2, 3, 4}\right\} \\
            & \\
            & \text{Calculate the base weight fractions of each component:} \\
            & \quad W_2 = (1 - W_1) \cdot G_1\\
            & \quad W_3 = 1.0 - W_1 - W_2 \\
            & \\
            & \text{if $W_1 \leq 0.5$:} \\
            & \quad \text{$P_{11}$ is given and W_xx is derived as} \\
            & \quad W_{xx} = W_{22} + W_{23} + W_{32} + W_{23} = W_1 \cdot (1 - P_{11}) + W_2 + W_3 \\ 
            & \\
            & \text{if $W_1 > 0.5$:} \\
            & \quad \text{$P_{xx}$ is given and $P_{11}$ is derived further down} \\
            & \quad W_{xx} = W_{22} + W_{23} + W_{32} + W_{23} = P_{xx} \cdot (W_2 + W_3) \\
            & \\
            & W_{22} = W_{xx} \cdot G_2 \cdot G_3 \\
            & W_{23} = W_{22} \cdot G_3^{-1} \\
            & W_{32} = W_{xx} \cdot (1 - G_2) \\
            & W_{33} = G_4^{-1} \cdot W_{32} \\
            & \\
            & P_{23} = 
            \begin{dcases}
                \dfrac{W_{23}}{W_2}, & \text{if $W_2 > 0$} \\
                0, & \text{otherwise}
            \end{dcases} \\
            & P_{12} = 1 - P_{22} - P_{23} \\
            & \\
            & P_{32} =
            \begin{dcases}
                \frac{W_{32}}{W_3}, & \text{if $W_3 > 0$} \\
                0, & \text{otherwise}
            \end{dcases} \\
            & P_{33} =
            \begin{dcases}
                \frac{W_{33}}{W_3}, & \text{if $W_3 > 0$} \\
                0, & \text{otherwise}
            \end{dcases} \\
            & P_{31} = 1 - P_{32} - P_{33} \\
            & \\
            & P_{12} =
            \begin{dcases}
                \frac{W_2 - W_{22} - W_{32}}{W_1}, & \text{if $W_1 > 0$} \\
                0, & \text{otherwise}
            \end{dcases} \\
            & P_{13} =
            \begin{dcases}
                \frac{W_3 - W_{23} - W_{33}}{W_1}, & \text{if $W_1 > 0$} \\
                0, & \text{otherwise}
            \end{dcases} \\
            & \\
            & \text{if $W_1 > 0.5$}: \\
            & \quad P_{11} = 1 - P_{12} - P_{13} \\
            & \\
            & \text{Remainder of weight fraction can be calculated as follows:} \\
            & \quad W_{ij} = {W_{ii}} \cdot {P_{ij}} \quad \forall {i,j} \in \left[ {1, 3} \right] \\
        \end{align*}
        
    """

    # MODEL METADATA:
    class Meta(_AbstractProbability.Meta):
        store_id = "R1G3Model"

    # PROPERTIES
    _G = 3

    inherit_W1 = BoolProperty(
        default=False, text="Inherit flag for W1",
        persistent=True, visible=True,
        set_action_name="update",
        mix_with=(SetActionMixin,)
    )
    W1 = FloatProperty(
        default=0.8, text="W1", math_text=r"$W_1$",
        persistent=True, visible=True, refinable=True, store_private=True,
        minimum=0.0, maximum=1.0, is_independent=True, inheritable=True,
        inherit_flag="inherit_W1", inherit_from="parent.based_on.probabilities.W1",
        set_action_name="update",
        mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
    )

    inherit_P11_or_P22 = BoolProperty(
        default=False, text="Inherit flag for P11_or_P22",
        persistent=True, visible=True,
        set_action_name="update",
        mix_with=(SetActionMixin,)
    )
    P11_or_P22 = FloatProperty(
        default=0.7, text="P11_or_P22",
        math_text=r"$P_{11} %s$ or $\newline P_{22} %s$" % (
            mt_range(0.0, "W_1", 0.5),
            mt_range(0.5, "W_1", 1.0)),
        persistent=True, visible=True, refinable=True, store_private=True,
        minimum=0.0, maximum=1.0, is_independent=True, inheritable=True,
        inherit_flag="inherit_P11_or_P22", inherit_from="parent.based_on.probabilities.P11_or_P22",
        set_action_name="update",
        mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
    )

    inherit_G1 = BoolProperty(
        default=False, text="Inherit flag for G1",
        persistent=True, visible=True,
        set_action_name="update",
        mix_with=(SetActionMixin,)
    )
    G1 = FloatProperty(
        default=0.7, text="W2/(W2+W3)",
        math_text=r"$\large\frac{W_2}{W_3 + W_2}$",
        persistent=True, visible=True, refinable=True, store_private=True,
        minimum=0.0, maximum=1.0, is_independent=True, inheritable=True,
        inherit_flag="inherit_G1", inherit_from="parent.based_on.probabilities.G1",
        set_action_name="update",
        mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
    )

    inherit_G2 = BoolProperty(
        default=False, text="Inherit flag for G2",
        persistent=True, visible=True,
        set_action_name="update",
        mix_with=(SetActionMixin,)
    )
    G2 = FloatProperty(
        default=0.7, text="(W22+W23)/(W22+W23+W32+W33)",
        math_text=r"$\large\frac{W_{22} + W_{23}}{W_{22} + W_{23} + W_{32} + W_{33}}$",
        persistent=True, visible=True, refinable=True, store_private=True,
        minimum=0.0, maximum=1.0, is_independent=True, inheritable=True,
        inherit_flag="inherit_G2", inherit_from="parent.based_on.probabilities.G2",
        set_action_name="update",
        mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
    )

    inherit_G3 = BoolProperty(
        default=False, text="Inherit flag for G3",
        persistent=True, visible=True,
        set_action_name="update",
        mix_with=(SetActionMixin,)
    )
    G3 = FloatProperty(
        default=0.7, text="W22/(W22+W23)",
        math_text=r"$\large\frac{W_{22}}{W_{22} + W_{23}}$",
        persistent=True, visible=True, refinable=True, store_private=True,
        minimum=0.0, maximum=1.0, is_independent=True, inheritable=True,
        inherit_flag="inherit_G3", inherit_from="parent.based_on.probabilities.G3",
        set_action_name="update",
        mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
    )

    inherit_G4 = BoolProperty(
        default=False, text="Inherit flag for G4",
        persistent=True, visible=True,
        set_action_name="update",
        mix_with=(SetActionMixin,)
    )
    G4 = FloatProperty(
        default=0.7, text="W23/(W32+W33)",
        math_text=r"$\large\frac{W_{22}}{W_{22} + W_{23}}$",
        persistent=True, visible=True, refinable=True, store_private=True,
        minimum=0.0, maximum=1.0, is_independent=True, inheritable=True,
        inherit_flag="inherit_G4", inherit_from="parent.based_on.probabilities.G4",
        set_action_name="update",
        mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
    )

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, W1=0.8, P11_or_P22=0.7, G1=0.7, G2=0.7, G3=0.7, G4=0.7,
            inherit_W1=False, inherit_P11_or_P22=False, inherit_G1=False,
            inherit_G2=False, inherit_G3=False, inherit_G4=False, *args, **kwargs):
        super(R1G3Model, self).__init__(R=1, *args, **kwargs)

        with self.data_changed.hold():
            self.W1 = not_none(W1, 0.8)
            self.inherit_W1 = bool(inherit_W1)
            self.P11_or_P22 = not_none(P11_or_P22, 0.7)
            self.inherit_P11_or_P22 = bool(inherit_P11_or_P22)
            self.G1 = not_none(G1, 0.7)
            self.inherit_G1 = bool(inherit_G1)
            self.G2 = not_none(G2, 0.7)
            self.inherit_G2 = bool(inherit_G2)
            self.G3 = not_none(G3, 0.7)
            self.inherit_G3 = bool(inherit_G3)
            self.G4 = not_none(G4, 0.7)
            self.inherit_G4 = bool(inherit_G4)

            self.update()

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def update(self):
        with self.monitor_changes():
            self.mW[0] = self.W1
            self.mW[1] = (1 - self.mW[0]) * self.G1
            self.mW[2] = 1.0 - self.mW[0] - self.mW[1]

            W0inv = 1.0 / self.mW[0] if self.mW[0] > 0.0 else 0.0

            Wxx = 0
            if self.mW[0] <= 0.5: # P00 given
                self.mP[0, 0] = self.P11_or_P22
                # Wxx = W11 + W12 + W21 + W22
                Wxx = self.mW[0] * (self.mP[0, 0] - 1) + self.mW[1] + self.mW[2]
            else: # Pxx given
                # Wxx = W11 + W12 + W21 + W22
                Wxx = (1.0 - self.mW[0]) * self.P11_or_P22

            # W11 + W12 = Wxx * G2:
            self.mW[1, 1] = Wxx * self.G2 * self.G3
            self.mW[1, 2] = Wxx * self.G2 * (1 - self.G3)
            self.mP[1, 1] = self.mW[1, 1] / self.mW[1] if self.mW[1] > 0.0 else 0.0

            self.mW[2, 1] = Wxx * (1 - self.G2) * self.G4
            self.mW[2, 2] = Wxx * (1 - self.G2) * (1 - self.G4)

            self.mP[1, 2] = (self.mW[1, 2] / self.mW[1]) if self.mW[1] > 0.0 else 0.0
            self.mP[1, 0] = 1 - self.mP[1, 1] - self.mP[1, 2]

            self.mP[2, 1] = (self.mW[2, 1] / self.mW[2]) if self.mW[2] > 0.0 else 0.0
            self.mP[2, 2] = (self.mW[2, 2] / self.mW[2]) if self.mW[2] > 0.0 else 0.0
            self.mP[2, 0] = 1 - self.mP[2, 1] - self.mP[2, 2]

            self.mP[0, 1] = (self.mW[1] - self.mW[1, 1] - self.mW[2, 1]) * W0inv
            self.mP[0, 2] = (self.mW[2] - self.mW[1, 2] - self.mW[2, 2]) * W0inv

            if self.mW[0] > 0.5:
                self.mP[0, 0] = 1 - self.mP[0, 1] - self.mP[0, 2]

            for i in range(3):
                for j in range(3):
                    self.mW[i, j] = self.mW[i, i] * self.mP[i, j]

            self.solve()
            self.validate()

    pass # end of class

@storables.register()
class R1G4Model(_AbstractProbability):

    r"""
    Probability model for Reichweite 1 with 4 components.
    
    The independent variables (# = g*(g-1) = 12) are:
    
    .. math::
        :nowrap:
        
            \begin{align*}
                & W_1
                & P_{11} (W_1 < 0,5)\text{ or }P_{xx} (W_1 > 0,5) 
                with P_{xx} = \frac {W_{22} + W_{23} + W_{24} + W_{32} + W_{33} + W_{34} + W_{42} + W_{43} + W_{44}}{W_2 + W_3 + W_4} \\
                & R_2 = \frac{ W_2 }{W_2 + W_3 + W_4}
                & R_3 = \frac{ W_3 }{W_3 + W_4} \\
                & G_2 = \frac{W_{22} + W_{23} + W_{24}}{\sum_{i=2}^{4}\sum_{j=2}^4{W_{ij}}}
                & G_3 = \frac{W_{32} + W_{33} + W_{34}}{\sum_{i=3}^{4}\sum_{j=2}^4{W_{ij}}} \\
                & G_{22} = \frac{W_{22}}{W_{22} + W_{23} + W_{24}}
                & G_{23} = \frac{W_{23}}{W_{23} + W_{24}} \\
                & G_{32} = \frac{W_{32}}{W_{32} + W_{33} + W_{34}}
                & G_{33} = \frac{W_{33}}{W_{33} + W_{34}} \\
                & G_{42} = \frac{W_{42}}{W_{42} + W_{43} + W_{44}}
                & G_{44} = \frac{W_{43}}{W_{43} + W_{44}}
            \end{align*} 
    
    Calculation of the other variables happens as follows:
    
    .. math::
        :nowrap:
        
        \begin{align*}
            & \text{Calculate the base weight fractions of each component:} \\
            & W_2 = (1 - W_1) \cdot R_1 \\
            & W_3 = (1 - W_1 - W_2) \cdot R_2 \\
            & W_4 = (1 - W_1 - W_2 - W_3) \\
            & \\
            & \text{if $W_1 \leq 0.5$:} \\
            & \quad \text{$P_{11}$ is given}\\
            & \quad W_{xx} = W_{22} + W_{23} + W_{24} + W_{32} + W_{33} + W_{34} + W_{42} + W_{43} + W_{44} = W_1 \cdot (1 - P_{11}) + W_2 + W_3 + W_4 \\
            & \text{if $W_1 > 0.5$:} \\
            & \quad \text{$P_{xx}$ is given and $P_{11}$ is derived further down} \\
            & \quad W_{xx} = W_{22} + W_{23} + W_{24} + W_{32} + W_{33} + W_{34} + W_{42} + W_{43} + W_{44} = P_{xx} \cdot (W_2 + W_3 + W_4) \\  
            & \\
            & \text{Caclulate a partial sum of the $2^{nd}$ component's contributions: } \\
            & W_{2x} = W_{xx} \cdot G_2 \\
            & \text{Calculate a partial sum of the $3^{d}$ and $4^{th}$ component's contributions:} \\
            & W_{yx} = W_{xx} - W_{2x} \\
            & \text{Calculate a partial sum of the $3^{d}$ component's contributions:} \\
            & W_{3x} = W_{yx} \cdot G_3 \\
            & \text{Calculate a partial sum of the $4^{th}$ component's contributions:} \\
            & W_{4x} = W_{yx} - W_{3x} \\
            & \\
            & W_{22} = G_{22} \cdot W_{2x} \\
            & W_{23} = G_{23} \cdot (W_{2x} - W_{22}) \\
            & W_{24} = W{2x} - W_{22} - W_{23} \\
            & \\
            & W_{32} = G_{32} \cdot W_{3x} \\
            & W_{33} = G_{33} \cdot (W_{3x} - W_{32}) \\
            & W_{34} = W{3x} - W_{32} - W_{33} \\
            & \\
            & W_{42} = G_{42} \cdot W_{4x} \\
            & W_{43} = G_{43} \cdot (W_{4x} - W_{42}) \\
            & W_{44} = W{4x} - W_{42} - W_{43} \\
            & \\
            & \text{ From the above weight fractions the junction probabilities 
                for any combination of $2^{nd}$, $3^{d}$ and $4^{th}$ type
                components can be calculated. } \\
            & \text{ The remaining probabilities are: } \\
            & P_{12} = \begin{dcases}
                \frac{W_2 - W_{22} - W_{32} - W_{42}}{W_1}, & \text{if $W_1 > 0$} \\
                0, & \text{otherwise}
            \end{dcases} \\
            & P_{13} = \begin{dcases}
                \frac{W_3 - W_{23} - W_{33} - W_{43}}{W_1}, & \text{if $W_1 > 0$} \\
                0, & \text{otherwise}
            \end{dcases} \\
            & P_{14} = \begin{dcases}
                \frac{W_4 - W_{24} - W_{34} - W_{44}}{W_1}, & \text{if $W_1 > 0$} \\
                0, & \text{otherwise}
            \end{dcases} \\
            & \\
            & \text{if $W_1 \leq 0.5$}: \\
            & \quad P_{11} = 1 - P_{12} - P_{13} - P_{14} \\
            & \text{Remainder of weight fraction can now be calculated as follows:} \\
            & \quad W_{ij} = {W_{ii}} \cdot {P_{ij}} \quad \forall {i,j} \in \left[ {1, 4} \right] \\
        \end{align*}
    """

    # MODEL METADATA:
    class Meta(_AbstractProbability.Meta):
        store_id = "R1G4Model"

    # PROPERTIES
    _G = 4

    inherit_W1 = BoolProperty(
        default=False, text="Inherit flag for W1",
        persistent=True, visible=True,
        set_action_name="update",
        mix_with=(SetActionMixin,)
    )
    W1 = FloatProperty(
        default=0.6, text="W1", math_text=r"$W_1$",
        persistent=True, visible=True, refinable=True, store_private=True,
        minimum=0.0, maximum=1.0, is_independent=True, inheritable=True,
        inherit_flag="inherit_W1", inherit_from="parent.based_on.probabilities.W1",
        set_action_name="update",
        mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
    )

    inherit_P11_or_P22 = BoolProperty(
        default=False, text="Inherit flag for P11_or_P22",
        persistent=True, visible=True,
        set_action_name="update",
        mix_with=(SetActionMixin,)
    )
    P11_or_P22 = FloatProperty(
        default=0.25, text="P11_or_P22",
        math_text=r"$P_{11} %s$ or $\newline P_{22} %s$" % (
            mt_range(0.0, "W_1", 0.5),
            mt_range(0.5, "W_1", 1.0)),
        persistent=True, visible=True, refinable=True, store_private=True,
        minimum=0.0, maximum=1.0, is_independent=True, inheritable=True,
        inherit_flag="inherit_P11_or_P22", inherit_from="parent.based_on.probabilities.P11_or_P22",
        set_action_name="update",
        mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
    )

    inherit_R1 = BoolProperty(
        default=False, text="Inherit flag for R1",
        persistent=True, visible=True,
        set_action_name="update",
        mix_with=(SetActionMixin,)
    )
    R1 = FloatProperty(
        default=0.5, text="W2/(W2+W3+W4)",
        math_text=r"$\large\frac{W_2}{W_2 + W_3 + W_4}$",
        persistent=True, visible=True, refinable=True, store_private=True,
        minimum=0.0, maximum=1.0, is_independent=True, inheritable=True,
        inherit_flag="inherit_R1", inherit_from="parent.based_on.probabilities.R1",
        set_action_name="update",
        mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
    )

    inherit_R2 = BoolProperty(
        default=False, text="Inherit flag for R2",
        persistent=True, visible=True,
        set_action_name="update",
        mix_with=(SetActionMixin,)
    )
    R2 = FloatProperty(
        default=0.5, text="W3/(W3+W4)",
        math_text=r"$\large\frac{W_3}{W_3 + W_4}$",
        persistent=True, visible=True, refinable=True, store_private=True,
        minimum=0.0, maximum=1.0, is_independent=True, inheritable=True,
        inherit_flag="inherit_R2", inherit_from="parent.based_on.probabilities.R2",
        set_action_name="update",
        mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
    )

    inherit_G1 = BoolProperty(
        default=False, text="Inherit flag for G1",
        persistent=True, visible=True,
        set_action_name="update",
        mix_with=(SetActionMixin,)
    )
    G1 = FloatProperty(
        default=0.5, text="(W22+W23+W24)/(W22+W23+W24+W32+W33+W34+W42+W43+W44)",
        math_text=r"$\large\frac{\sum_{j=2}^{4} W_{2j}}{\sum_{i=2}^{4} \sum_{j=2}^{4} W_{ij}}$",
        persistent=True, visible=True, refinable=True, store_private=True,
        minimum=0.0, maximum=1.0, is_independent=True, inheritable=True,
        inherit_flag="inherit_G1", inherit_from="parent.based_on.probabilities.G1",
        set_action_name="update",
        mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
    )

    inherit_G2 = BoolProperty(
        default=False, text="Inherit flag for G2",
        persistent=True, visible=True,
        set_action_name="update",
        mix_with=(SetActionMixin,)
    )
    G2 = FloatProperty(
        default=0.4, text="(W32+W33+W34)/(W32+W33+W34+W42+W43+W44)",
        math_text=r"$\large\frac{\sum_{j=2}^{4} W_{3j}}{\sum_{i=3}^{4} \sum_{j=2}^{4} W_{ij}}$",
        persistent=True, visible=True, refinable=True, store_private=True,
        minimum=0.0, maximum=1.0, is_independent=True, inheritable=True,
        inherit_flag="inherit_G2", inherit_from="parent.based_on.probabilities.G2",
        set_action_name="update",
        mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
    )

    inherit_G11 = BoolProperty(
        default=False, text="Inherit flag for G11",
        persistent=True, visible=True,
        set_action_name="update",
        mix_with=(SetActionMixin,)
    )
    G11 = FloatProperty(
        default=0.5, text="W22/(W22+W23+W24)",
        math_text=r"$\large\frac{W_{22}}{\sum_{j=2}^{4} W_{2j}}$",
        persistent=True, visible=True, refinable=True, store_private=True,
        minimum=0.0, maximum=1.0, is_independent=True, inheritable=True,
        inherit_flag="inherit_G11", inherit_from="parent.based_on.probabilities.G11",
        set_action_name="update",
        mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
    )

    inherit_G12 = BoolProperty(
        default=False, text="Inherit flag for G12",
        persistent=True, visible=True,
        set_action_name="update",
        mix_with=(SetActionMixin,)
    )
    G12 = FloatProperty(
        default=0.5, text="W23/(W23+W24)",
        math_text=r"$\large\frac{W_{23}}{\sum_{j=3}^{4} W_{2j}}$",
        persistent=True, visible=True, refinable=True, store_private=True,
        minimum=0.0, maximum=1.0, is_independent=True, inheritable=True,
        inherit_flag="inherit_G12", inherit_from="parent.based_on.probabilities.G12",
        set_action_name="update",
        mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
    )

    inherit_G21 = BoolProperty(
        default=False, text="Inherit flag for G21",
        persistent=True, visible=True,
        set_action_name="update",
        mix_with=(SetActionMixin,)
    )
    G21 = FloatProperty(
        default=0.8, text="W32/(W32+W33+W34)",
        math_text=r"$\large\frac{W_{32}}{\sum_{j=2}^{4} W_{3j}}$",
        persistent=True, visible=True, refinable=True, store_private=True,
        minimum=0.0, maximum=1.0, is_independent=True, inheritable=True,
        inherit_flag="inherit_G21", inherit_from="parent.based_on.probabilities.G21",
        set_action_name="update",
        mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
    )

    inherit_G22 = BoolProperty(
        default=False, text="Inherit flag for G22",
        persistent=True, visible=True,
        set_action_name="update",
        mix_with=(SetActionMixin,)
    )
    G22 = FloatProperty(
        default=0.8, text="W33/(W32+W34)",
        math_text=r"$\large\frac{W_{33}}{\sum_{j=3}^{4} W_{3j}}$",
        persistent=True, visible=True, refinable=True, store_private=True,
        minimum=0.0, maximum=1.0, is_independent=True, inheritable=True,
        inherit_flag="inherit_G22", inherit_from="parent.based_on.probabilities.G22",
        set_action_name="update",
        mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
    )

    inherit_G31 = BoolProperty(
        default=False, text="Inherit flag for G31",
        persistent=True, visible=True,
        set_action_name="update",
        mix_with=(SetActionMixin,)
    )
    G31 = FloatProperty(
        default=0.7, text="W42/(W42+W43+W44)",
        math_text=r"$\large\frac{W_{42}}{\sum_{j=2}^{4} W_{4j}}$",
        persistent=True, visible=True, refinable=True, store_private=True,
        minimum=0.0, maximum=1.0, is_independent=True, inheritable=True,
        inherit_flag="inherit_G31", inherit_from="parent.based_on.probabilities.G31",
        set_action_name="update",
        mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
    )

    inherit_G32 = BoolProperty(
        default=False, text="Inherit flag for G32",
        persistent=True, visible=True,
        set_action_name="update",
        mix_with=(SetActionMixin,)
    )
    G32 = FloatProperty(
        default=0.5, text="W43/(W43+W44)",
        math_text=r"$\large\frac{W_{43}}{\sum_{j=3}^{4} W_{4j}}$",
        persistent=True, visible=True, refinable=True, store_private=True,
        minimum=0.0, maximum=1.0, is_independent=True, inheritable=True,
        inherit_flag="inherit_G32", inherit_from="parent.based_on.probabilities.G32",
        set_action_name="update",
        mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
    )

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------

    def __init__(self, W1=0.6, P11_or_P22=0.25, R1=0.5, R2=0.5, G1=0.5, G2=0.4,
            G11=0.5, G12=0.5, G21=0.8, G22=0.75, G31=0.7, G32=0.5,
            inherit_W1=False, inherit_P11_or_P22=False, inherit_R1=False,
            inherit_R2=False, inherit_G1=False, inherit_G2=False,
            inherit_G11=False, inherit_G12=False, inherit_G21=False,
            inherit_G22=False, inherit_G31=False, inherit_G32=False, *args, **kwargs):
        super(R1G4Model, self).__init__(R=1, *args, **kwargs)

        with self.data_changed.hold():
            self.W1 = not_none(W1, 0.6)
            self.inherit_W1 = inherit_W1
            self.P11_or_P22 = not_none(P11_or_P22, 0.25)
            self.inherit_P11_or_P22 = inherit_P11_or_P22
            self.R1 = not_none(R1, 0.5)
            self.inherit_R1 = inherit_R1
            self.R2 = not_none(R2, 0.5)
            self.inherit_R2 = inherit_R2
            self.G1 = not_none(G1, 0.5)
            self.inherit_G1 = inherit_G1
            self.G2 = not_none(G2, 0.4)
            self.inherit_G2 = inherit_G2
            self.G11 = not_none(G11, 0.5)
            self.inherit_G11 = inherit_G11
            self.G12 = not_none(G12, 0.5)
            self.inherit_G12 = inherit_G12
            self.G21 = not_none(G21, 0.8)
            self.inherit_G21 = inherit_G21
            self.G22 = not_none(G22, 0.75)
            self.inherit_G22 = inherit_G22
            self.G31 = not_none(G31, 0.7)
            self.inherit_G31 = inherit_G31
            self.G32 = not_none(G32, 0.5)
            self.inherit_G32 = inherit_G32

            self.update()

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def update(self):
        with self.monitor_changes():
            self.mW[0] = self.W1
            self.mW[1] = (1.0 - self.mW[0]) * self.R1
            self.mW[2] = (1.0 - self.mW[0] - self.mW[1]) * self.R2
            self.mW[3] = 1.0 - self.mW[0] - self.mW[1] - self.mW[2]

            W0inv = 1.0 / self.mW[0] if self.mW[0] > 0.0 else 0.0

            if self.mW[0] < 0.5: # P11 is given
                self.mP[0, 0] = self.P11_or_P22
                Wxx = self.mW[0] * (self.mP[0, 0] - 1) + self.mW[1] + self.mW[2] + self.mW[3]
            else: # P22 is given
                Wxx = self.P11_or_P22 * (self.mW[1] + self.mW[2] + self.mW[3])

            W1x = Wxx * self.G1 # = W11 + W12 + W13
            Wyx = (Wxx - W1x)   # = W21 + W22 + W23 + W31 + W32 + W33
            W2x = Wyx * self.G2 # = W21 + W22 + W23
            W3x = Wyx - W2x     # = W31 + W32 + W33

            self.mW[1, 1] = self.G11 * W1x
            self.mW[1, 2] = self.G12 * (W1x - self.mW[1, 1])
            self.mW[1, 3] = W1x - self.mW[1, 1] - self.mW[1, 2]

            self.mW[2, 1] = self.G21 * W2x
            self.mW[2, 2] = self.G22 * (W2x - self.mW[2, 1])
            self.mW[2, 3] = W2x - self.mW[2, 1] - self.mW[2, 2]

            self.mW[3, 1] = self.G31 * W3x
            self.mW[3, 2] = self.G32 * (W3x - self.mW[3, 1])
            self.mW[3, 3] = W3x - self.mW[3, 1] - self.mW[3, 2]

            for i in range(1, 4):
                self.mP[i, 0] = 1
                for j in range(1, 4):
                    self.mP[i, j] = self.mW[i, j] / self.mW[i] if self.mW[i] > 0 else 0
                    self.mP[i, 0] -= self.mP[i, j]
                self.mW[i, 0] = self.mW[i] * self.mP[i, 0]

            self.mP[0, 1] = (self.mW[1] - self.mW[1, 1] - self.mW[2, 1] - self.mW[3, 1]) * W0inv
            self.mP[0, 2] = (self.mW[2] - self.mW[1, 2] - self.mW[2, 2] - self.mW[3, 2]) * W0inv
            self.mP[0, 3] = (self.mW[3] - self.mW[1, 3] - self.mW[2, 3] - self.mW[3, 3]) * W0inv

            if self.mW[0] >= 0.5:
                self.mP[0, 0] = 1 - self.mP[0, 1] - self.mP[0, 2] - self.mP[0, 3]

            for i in range(4):
                for j in range(4):
                    self.mW[i, j] = self.mW[i] * self.mP[i, j]

            self.solve()
            self.validate()

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import numpy as np 

class RefineHistory(object):
    """
    A history tracking class for refinements
    """
    
    INITIAL_ITERATION_INDEX = -1
    LAST_ITERATION_INDEX = -1
    
    ITERATION_INDEX = 0
    RESIDUAL_INDEX = -1
    SOLUTION_SELECTOR = np.s_[ITERATION_INDEX+1:RESIDUAL_INDEX]
    PLOT_SAMPLE_SELECTOR = np.s_[ITERATION_INDEX+1:]
    
    samples = None
    
    _closed = False
    
    @property
    def best_entry(self):
        if self._closed:
            samples = self.samples
        else:
            samples = np.asanyarray(self.samples)
        residuals = samples[:,self.RESIDUAL_INDEX]
        return samples[np.where(residuals == residuals.min()),:][-1].flatten()
    
    @property
    def initial_entry(self):
        if self._closed:
            samples = self.samples
        else:
            samples = np.asanyarray(self.samples)
        iterations = samples[:,self.ITERATION_INDEX]       
        return samples[np.where(iterations == self.INITIAL_ITERATION_INDEX),:][-1].flatten()
    
    @property
    def last_entry(self):
        if self._closed:
            samples = self.samples
        else:
            samples = np.asanyarray(self.samples)
        iterations = samples[:,self.ITERATION_INDEX]
        return samples[np.where(iterations == self.LAST_ITERATION_INDEX),:][-1].flatten()
    
    @property
    def best_solution(self):
        return self.best_entry[self.SOLUTION_SELECTOR].flatten()
    
    @property
    def initial_solution(self):
        return self.initial_entry[self.SOLUTION_SELECTOR]
    
    @property
    def last_solution(self):
        return self.last_entry[self.SOLUTION_SELECTOR]

    @property
    def initial_residual(self):
        return float(self.initial_entry[self.RESIDUAL_INDEX])

    @property
    def best_residual(self):
        return float(self.best_entry[self.RESIDUAL_INDEX])
    
    @property
    def last_residual(self):
        return float(self.last_entry[self.RESIDUAL_INDEX])

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self):
        self.samples = []

    def _sort_solutions_by_iteration(self):
        self.samples.sort(key=lambda s: s[0])
    
    # ------------------------------------------------------------
    #      ContextManager implementation
    # ------------------------------------------------------------    
    def close(self):
        self._closed = True
        self._sort_solutions_by_iteration()
        self.samples = np.asanyarray(self.samples)
    
    def __enter__(self):
        assert self._closed is False, "Cannot use a closed refinement history!"
        return self
    
    def __exit__(self, tp, value, traceback):
        self.close()

    # ------------------------------------------------------------
    #      Solution registration:
    # ------------------------------------------------------------    
    def set_initial_solution(self, solution, residual):
        if self._closed:
            raise RuntimeError("Cannot change a closed refinement history!")
        self.register_solution(self.INITIAL_ITERATION_INDEX, solution, residual)
        
    def register_solution(self, iteration, solution, residual):
        if self._closed:
            raise RuntimeError("Cannot change a closed refinement history!")
        sample = [iteration,]+list(solution)+[residual,]
        self.samples.append(sample)
        if iteration > self.LAST_ITERATION_INDEX:
            self.LAST_ITERATION_INDEX = iteration
            
    def get_residual_per_iteration(self):
        if not self._closed:
            raise RuntimeError("Cannot perform analysis on an open refinement history")
        return self.samples[:,[self.ITERATION_INDEX,self.RESIDUAL_INDEX]].tolist()
        
    pass #end of class
        


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import time

from mvc.support.utils import not_none  # @UnusedImport but keep it as an alias

def u(string):
    if isinstance(string, bytes):
        return string.decode(errors='replace')
    else:
        return string

def print_timing(func):
    def wrapper(*args, **kwargs):
        t1 = time.time()
        res = func(*args, **kwargs)
        t2 = time.time()
        print('%s took %0.3f ms' % (func.__name__, (t2 - t1) * 1000.0))
        return res
    return wrapper


# coding=UTF-8
# ex:ts=4:sw=4:et=on
import time

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

class RefineStatus(object):
    """
    A status object for a refinement. This should provide some hints
    for the UI to keep track of the status of the refinement.
    
    The error, running and finished flags are mutually exclusive but can be
    False together. When the cancelled flag is set, the user cancelled the
    refinement (the stop signal was set). When the error flag is true, an error
    was encountered during the refinement. When the finished flag is true the 
    refinement has finished successfully. When the running flag is true the 
    refinement is still running. When all three flags are False, the refinement 
    has not started yet.
    
    The status message should be a textual description of these three flags,
    but can be set to any value.
    
    The current error is retrieved from the RefinementHistory instance passed
    in the constructor (it is the residual of the last solution registered).
    
    The best way to use the status object is with a context, like this:
    
     with RefineHistory() as history:
         with RefineStatus(history) as status:
             run_refinement()
             
    """
    
    _error = False
    @property
    def error(self):
        return self._error
    @error.setter
    def error(self, value):
        self._error = bool(value)
        if self._error: 
            self.running = False
            self.cancelled = False
            self.finished = False
    
    _cancelled = False
    @property
    def cancelled(self):
        return self._cancelled
    @cancelled.setter
    def cancelled(self, value):
        self._cancelled = bool(value)
        if self._cancelled: 
            self.running = False
            self.error = False
            self.finished = False
    
    _running = False
    @property
    def running(self):
        return self._running
    @running.setter
    def running(self, value):
        self._running = bool(value)
        if self._running: 
            self.error = False
            self.cancelled = False
            self.finished = False
        
    _finished = False
    @property
    def finished(self):
        return self._finished
    @finished.setter
    def finished(self, value):
        self._finished = bool(value)
        if self._finished:
            self.error = False
            self.cancelled = False
            self.running = False
     
    message = "Not initialized."
    
    @property
    def current_error(self):
        return self.history.last_solution[self.history.RESIDUAL_INDEX]
    
    def __init__(self, history, stop_signal=None):
        assert history is not None, "The RefinementStatus needs a RefinementHistory instance!"
        self.history = history
        self.stop_signal = stop_signal
        self.message = "Initialized."
        
        self.start_time = -1
        self.end_time = -1
    
    def __enter__(self):
        # Set flag
        self.running = True
        # Set message
        self.message = "Running..."
        # Record start time
        self.start_time = time.time()
        # Return us
        return self
        
    def __exit__(self, tp, value, traceback):
        # Record end time
        self.start_time = -1

        if tp is not None:
            self.message = "Refinement error!"
            self.error = True
        else:
            if self.stop_signal is not None and self.stop_signal.is_set():            
                self.message = "Refinement cancelled!"
                self.cancelled = True
            else:
                self.message = "Refinement finished!"
                self.finished = True
    
    def get_total_time(self):
        """ Gets the total time the refinement has run in ms """
        if self.start_time == -1:
            return 0
        else:
            return (self.end_time - self.start_time) * 1000.0
    
    pass #end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

class RefinableMixin(object):
    """
    Mixing for the ~:class:`mvc.models.properties.LabeledProperty` descriptor
    that allows the property to be refinable.
    When this Mixin is used, the user should pass 4 additional keyword 
    arguments to the descriptor:
        - refinable: boolean set to True if the property is refinable
        - refinable_info_format: the format for the refinement info attribute
        - minimum: the minimum allowed value (or None as default)
        - maximum: the maximum allowed value (or None as default) 
    """

    refinable = True
    refinable_info_format = "%(label)s_ref_info"

    minimum = None
    maximum = None

    def get_refinement_info_name(self):
        return self.refinable_info_format % { 'label': self.label }

    pass #end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from collections import OrderedDict
import logging
logger = logging.getLogger(__name__)

from io import StringIO

from zipfile import ZipFile
try:
    # Check if zlib is available, if so, we can use compression when saving
    import zlib #@UnusedImport
    from zipfile import ZIP_DEFLATED as COMPRESSION
except ImportError:
    from zipfile import ZIP_STORED as COMPRESSION

from ..utils import not_none
from .json_codec import PyXRDDecoder, PyXRDEncoder

class StorableRegistry(dict):
    """
        Basically a dict which maps class names to the actual
        class types. This relies on the classes being registered using
        the 'register' decorator provided in this class type.
        It also has a number of aliases, for backwards-compatibility (e.g.
        when class names change or from the era before using this method
        when we stored the entire class path, we might remove these at
        some point)
    """

    # For backwards compatibility:
    aliases = {
        'generic.treemodels/XYListStore': 'XYListStore',
        'generic.treemodels/ObjectListStore': 'ObjectListStore',
        'generic.treemodels/ObjectTreeStore': 'ObjectTreeStore',
        'generic.treemodels/IndexListStore': 'IndexListStore',
        'generic.models.treemodels/ObjectListStore': 'ObjectListStore',
        'generic.models.treemodels/ObjectTreeStore': 'ObjectTreeStore',
        'generic.models.treemodels/IndexListStore': 'IndexListStore',
        'generic.models.treemodels/XYListStore': 'XYListStore',
        'generic.models/PyXRDLine': 'PyXRDLine',
        'generic.models/CalculatedLine': 'CalculatedLine',
        'generic.models/ExperimentalLine': 'ExperimentalLine',
        'goniometer.models/Goniometer': 'Goniometer',
        'specimen.models/Specimen': 'Specimen',
        'specimen.models/Marker': 'Marker',
        'mixture.models/Mixture': 'Mixture',
        'atoms.models/AtomType': 'AtomType',
        'atoms.models/Atom': 'Atom',
        'probabilities.R0models/R0G1Model': 'R0G1Model',
        'probabilities.R0models/R0G2Model': 'R0G2Model',
        'probabilities.R0models/R0G3Model': 'R0G3Model',
        'probabilities.R0models/R0G4Model': 'R0G4Model',
        'probabilities.R0models/R0G5Model': 'R0G5Model',
        'probabilities.R0models/R0G6Model': 'R0G6Model',
        'probabilities.R1models/R1G2Model': 'R1G2Model',
        'probabilities.R1models/R1G3Model': 'R1G3Model',
        'probabilities.R1models/R1G4Model': 'R1G4Model',
        'probabilities.R2models/R2G2Model': 'R2G2Model',
        'probabilities.R2models/R2G3Model': 'R2G3Model',
        'probabilities.R3models/R3G2Model': 'R3G2Model',
        'phases.CSDS_models/LogNormalCSDSDistribution': 'LogNormalCSDSDistribution',
        'phases.CSDS_models/DritsCSDSDistribution': 'DritsCSDSDistribution',
        'phases.atom_relations/AtomRelation': 'AtomRelation',
        'phases.atom_relations/AtomRatio': 'AtomRatio',
        'phases.atom_relations/AtomContents': 'AtomContents',
        'phases.models/UnitCellProperty': 'UnitCellProperty',
        'phases.models/Component': 'Component',
        'phases.models/Phase': 'Phase',
        'project.models/Project': 'Project',
        'InSituMixture': None
    }

    def __getitem__(self, key):
        key = self.aliases.get(key, key)
        return super(StorableRegistry, self).__getitem__(key) if key is not None else None

    def register(self):
        """
            Returns a decorator that will register Storable sub-classes.
        """
        return self.register_decorator

    def register_decorator(self, cls):
        if hasattr(cls, 'Meta') and hasattr(cls.Meta, 'store_id'):
            logger.debug("Registering %s as storage type with id '%s'" % (cls, cls.Meta.store_id))
            self[cls.Meta.store_id] = cls
        else:
            raise TypeError("Cannot register type '%s' without a Meta.store_id!" % cls)
        return cls

    pass # end of class

# This is filled using register decorator
storables = StorableRegistry()

# Needs to be importable, could be used for more compact Python pickling:
def __map_reduce__(json_obj):
    decoder = PyXRDDecoder(mapper=storables)
    return decoder.decode(json_obj)

class Storable(object):
    """
        A class with a number of default implementations to serialize objects
        to JSON strings. It used the PyXRDDecoder en PyXRDEncoder.
        Subclasses should override their 'Meta.store_id' property
        and register themselves by calling the storables.register method
        and applying it as decorator to the subclass:
        
         @storables.register()
         class StorableSubclass(Storable, ...):
            ...
            
        Sub-classes can optionally implement the following methods:
         - 'json_properties' or for more fine-grained control 'to_json'
         - 'from_json'
         
    """
    __storables__ = []

    class Meta():  # Sub classes need to override this and set store_id!!
        store_id = None

    ###########################################################################
    # High-level JSON (de)serialisiation related methods & functions:
    ###########################################################################
    def dump_object(self, zipped=False):
        """
            Returns this object serialized as a JSON string.
            If `zipped` is true it returns an in-memory ZipFile.
        """
        content = PyXRDEncoder.dump_object(self)
        if zipped:
            f = StringIO()
            with ZipFile(f, mode="w", compression=COMPRESSION) as z:
                z.writestr('content', content)
            return f
        else:
            return content

    def print_object(self):
        """
        Prints the output from dump_object().
        """
        print(self.dump_object())

    ###########################################################################
    # Low-level JSON (de)serialisiation related methods & functions:
    ###########################################################################
    def to_json(self):
        """
        Method that should return a dict containing two keys:
         - 'type' -> registered class Meta.store_id
         - 'properties' -> a dict containg all the properties neccesary to 
           re-create the object when serialized as JSON.
        """
        return {
            "type": self.Meta.store_id,
            "properties": self.json_properties()
        }

    def to_json_multi_part(self):
        """
            This should generate a list of two-tuples:
            (partname, json_dict), (partname, json_dict), ...
            These can then be saved as seperate files (e.g. in a ZIP file)
        """
        yield ('content', self.to_json())

    def json_properties(self):
        """
        Method that should return a dict containing all the properties necessary to 
        re-create the object when serialized as JSON.
        """
        retval = OrderedDict()
        def add_prop(label, store_private):
            if not store_private:
                retval[label] = getattr(self, label)
            else:
                try:
                    retval[label] = getattr(self, store_private)
                except (TypeError, AttributeError):
                    retval[label] = getattr(type(self), label)._get(self)

        from mvc.models import Model
        if isinstance(self, Model):
            for prop in self.Meta.all_properties:
                if prop.persistent:
                    add_prop(prop.label, not_none(prop.store_private, False))
                    if getattr(prop, 'refinable', False):
                        add_prop(prop.get_refinement_info_name(), False)
        elif hasattr(self, "__storables__"): # Fallback:
            for val in self.__storables__:
                add_prop(val, False)
        else:
            raise RuntimeError("Cannot find either a '__storables__' or Meta class attribute on Storable '%s' instance!" % type(self))
        return retval

    def parse_init_arg(self, arg, default, child=False, default_is_class=False, **kwargs):
        """
        Can be used to transform an argument passed to a __init__ method of a
        Storable (sub-)class containing a JSON dict into the actual object it
        is representing.
        
        *arg* the passed argument
        
        *default* the default value if argument is None
        
        **child* boolean flag indicating wether or not the object is a child,
        if true, self is passed as the parent keyword to the JSON decoder if
        the passed argument is a JSON dict
        
        **default_is_class** boolean flag indicating whether or not the passed
        default value is an unitialized type. If True, the type will be initialized
        using the kwargs passed to this function.
        
        **kwargs* any other kwargs are passed to the JSON decoder if the passed
        argument is a JSON dict
        
        :rtype: the argument (not a JSON dict), the actual object (argument was
        a JSON dict) or the default value (argument was None)
        """
        if arg == None:
            if not default_is_class:
                return default
            else:
                if child: kwargs["parent"] = self
                return default(**kwargs)
        elif isinstance(arg, dict) and "type" in arg and "properties" in arg:
            arg = PyXRDDecoder(mapper=storables, parent=self if child else None).__pyxrd_decode__(arg, **kwargs)
            return arg
        else:
            return arg

    @classmethod
    def from_json(cls, *args, **kwargs):
        """
        Class method transforming JSON kwargs into an instance of this class.
        By default this assumes a 1-on-1 mapping to the __init__ method.
        """
        return cls(*args, **kwargs)

    ###########################################################################
    # Others:
    ###########################################################################
    def __reduce__(self):
        props = self.dump_object()
        return __map_reduce__, (props,), None

    pass # end of class



from pyxrd.refinement.refine_method_meta import RefineMethodMeta

from .methods import *  # @UnusedWildImport

class RefineMethodManager(object):
    
    @classmethod
    def initialize_methods(cls, refine_options):
        """
            Returns a dict of refine methods as values and their index as key
            with the passed refine_options dict applied.
        """
        # 1. Create a list of refinement instances:
        refine_methods = {}
        for index, method in cls.get_all_methods().items():
            refine_methods[index] = method()

        # 2. Create dict of default options
        default_options = {}
        for method in list(refine_methods.values()):
            default_options[method.index] = {
                name: getattr(type(method), name).default for name in method.options
            }

        # 3. Apply the refine options to the methods
        if not refine_options == None:
            for index, options in zip(list(refine_options.keys()), list(refine_options.values())):
                index = int(index)
                if index in refine_methods:
                    method = refine_methods[index]
                    for arg, value in zip(list(options.keys()), list(options.values())):
                        if hasattr(method, arg):
                            setattr(method, arg, value)

        return refine_methods
    
    @classmethod
    def get_all_methods(cls):
        """ Returns all the registered refinement methods """
        return RefineMethodMeta.registered_methods
    
    @classmethod
    def get_method_from_index(cls, index):
        """ Returns the actual refinement method defined by the index """
        return cls.get_all_methods()[index]
    
    pass # end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from pkg_resources import resource_filename # @UnresolvedImport

from pyxrd.generic.views import DialogView

class AddInSituBehaviourView(DialogView):
    title = "Add Behaviour"
    subview_builder = resource_filename(__name__, "glade/add_behaviour.glade")
    subview_toplevel = "add_behaviour_container"

    def __init__(self, *args, **kwargs):
        DialogView.__init__(self, *args, **kwargs)

    def get_behaviour_type(self):
        itr = self.behaviour_combo_box.get_active_iter()
        val = self.behaviour_combo_box.get_model().get_value(itr, 1) if itr else None
        return val

    @property
    def behaviour_combo_box(self):
        return self["cmb_behaviours"]

    pass # end of class



import Pyro4
Pyro4.config.SERIALIZERS_ACCEPTED = ["json", "marshal", "serpent", "pickle"]
Pyro4.config.SERIALIZER = "pickle"
Pyro4.config.PICKLE_PROTOCOL_VERSION = 3
Pyro4.config.COMPRESSION = True
Pyro4.config.SERVERTYPE = "multiplex"
Pyro4.config.COMMTIMEOUT = 3.5
Pyro4.config.REQUIRE_EXPOSE = False
Pyro4.config.SOCK_REUSE = True

import platform
if platform.system() == "Windows" and float(platform.release()) >= 6:
    USE_MSG_WAITALL = True

PYRO_NAME = "pyxrd.server"
KEEP_SERVER_ALIVE = False # setting this to false may produce unwanted results!


# coding=UTF-8
# ex:ts=4:sw=4:et

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gdk

from .event_delegator import MPLCanvasEventDelegator

class EyeDropper():
    """
        A wrapper that makes eye-dropping a plot possible. Will call
        the click_callback with an x position and the click event object as
        arguments.
    """
    def __init__(self, plot_controller, click_callback=None):
        self._canvas = plot_controller.canvas
        self._window = self._canvas.get_window()
        self._click_callback = click_callback
        self.connect()

    def _on_motion(self, event):
        if self._window is not None:
            self._window.set_cursor(Gdk.Cursor.new(Gdk.CursorType.CROSSHAIR)) # @UndefinedVariable
        return False

    def _on_click(self, event):
        x_pos = -1
        if event.inaxes:
            x_pos = event.xdata
        if callable(self._click_callback):
            self._click_callback(x_pos, event)
        if self._window is not None:
            self._window.set_cursor(None)
        return True

    def connect(self):
        delegator = MPLCanvasEventDelegator.wrap_canvas(self._canvas)

        delegator.connect('motion_notify_event', self._on_motion, first=True)
        delegator.connect('button_press_event', self._on_click, first=True)

    def disconnect(self):
        if self._window is not None:
            self._window.set_cursor(None)

        delegator = MPLCanvasEventDelegator.wrap_canvas(self._canvas)
        delegator.disconnect('motion_notify_event', self._on_motion)
        delegator.disconnect('button_press_event', self._on_click)

    pass #end of class



#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import unittest

from test.utils import create_object_attribute_test

from pyxrd.phases.models.atom_relations import AtomRatio

__all__ = [
    'TestAtomRatio',
]

class DummyHoldableSignal():
    def ignore(self):
        return self
    def __enter__(self):
        pass
    def __exit__(self, *args):
        pass

class DummyAtom():
    data_changed = DummyHoldableSignal()
    attribute = 0.0

class DummyParent():
    inherit_atom_relations = False

class TestAtomRatio(unittest.TestCase):

    phase = None

    def setUp(self):
        self.atom1 = DummyAtom()
        self.atom2 = DummyAtom()
        self.parent = DummyParent()
        self.atom_ratio = AtomRatio(
            name="TestRatio",
            sum=2,
            value=0.5,
            atom1=[self.atom1, "attribute"],
            atom2=[self.atom2, "attribute"],
            parent=self.parent
        )
        self.atom_ratio.resolve_relations()

    def tearDown(self):
        del self.atom1
        del self.atom2
        del self.atom_ratio

    def test_not_none(self):
        self.assertIsNotNone(self.atom_ratio)
        self.assertIsNotNone(self.atom_ratio.atom1[0])
        self.assertIsNotNone(self.atom_ratio.atom2[0])

    def test_apply_relation(self):
        self.atom_ratio.enabled = True
        self.atom_ratio.apply_relation()
        self.assertEqual(self.atom1.attribute, 1.0)
        self.assertEqual(self.atom2.attribute, 1.0)
        self.atom_ratio.value = 0.1
        self.atom_ratio.apply_relation()
        self.assertEqual(self.atom1.attribute, 0.2)
        self.assertEqual(self.atom2.attribute, 1.8)

    test_name = create_object_attribute_test("atom_ratio", "name", "Test Name")
    test_name = create_object_attribute_test("atom_ratio", "value", 0.5)
    test_name = create_object_attribute_test("atom_ratio", "sum", 6)
    test_name = create_object_attribute_test("atom_ratio", "atom1", (None, "Test"))
    test_name = create_object_attribute_test("atom_ratio", "atom2", (None, "Test"))

    pass # end of class



#!/usr/bin/python
# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import shutil
import os
import tempfile

import unittest

from pyxrd.generic.exceptions import AlreadyRegistered, NotRegistered
from pyxrd.generic.io.data_registry import DataRegistry

__all__ = [
    'TestDataRegistry',
]


class TestDataRegistry(unittest.TestCase):

    def setUp(self):

        self.base_temp_dir = tempfile.mkdtemp()

        self.dirs = [
            ("TEST1", os.path.join(self.base_temp_dir, "test1/"), None),
            ("TEST2", "test2/", "TEST1"),
        ]
        self.files = [
            ("ROOTFILE", "root.txt", None),
            ("TEST1_FILE", "test1file.txt", "TEST1"),
            ("TEST2_FILE", "test2file.txt", "TEST2"),
        ]
        self.data_reg = DataRegistry(dirs=self.dirs, files=self.files)
        pass

    def tearDown(self):
        shutil.rmtree(self.base_temp_dir)
        del self.data_reg
        pass

    def test_exceptions(self):
        with self.assertRaises(NotRegistered):
            self.data_reg.get_directory_path("TEST3")

        with self.assertRaises(AlreadyRegistered):
            self.data_reg.register_data_directory("TEST1", "false_test1/", None)

    def test_iter(self):
        for path in self.data_reg.get_all_files():
            self.assertNotIn(path, [None, ""])

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from math import sqrt
import numpy as np
from .math_tools import smooth

def R_squared(exp, calc, *args):
    """
        Calculate the 'coefficient of determination' (aka R²) for the given
        experimental and calculated patterns (1D numpy arrays, y-values only) 
    """
    avg = sum(exp) / exp.size
    sserr = np.sum((exp - calc) ** 2)
    sstot = np.sum((exp - avg) ** 2)
    return 1 - (sserr / sstot)

def Rp(exp, calc, *args):
    """
        Calculate the 'pattern R factor' (aka Rp) for the given experimental and
        calculated patterns (1D numpy arrays, y-values only)
    """
    return np.sum(np.abs(exp - calc)) / np.sum(np.abs(exp)) * 100

def smooth_pattern(pattern):
    """
        Smooth the given pattern.
    """
    return smooth(pattern, 15)

def derive(pattern):
    """
        Calculate the first derivative pattern pattern. 
        Smoothes the input first, so noisy patterns shouldn't
        be much of a problem.
    """
    return np.gradient(smooth_pattern(pattern))

def Rpder(exp, calc):
    """
        Calculated the 'derived pattern R factor' (aka Rp') for the given 
        experimental and calculated patterns.
    """
    return Rp(derive(exp), derive(calc))

def Rpw(exp, calc):
    """
        Calculates the 'weighted pattern R factor' (aka Rwp) for the given experimental and
        calculated patterns (1D numpy arrays, y-values only)
        The weights are set equal to the inverse of the observed intensities.  
    """
    # weighted Rp:
    # Rwp = Sqrt ( Sum[w * (obs - calc)²] / Sum[w * obs²] )  w = 1 / Iobs
    sm1 = 0
    sm2 = 0
    for i in range(exp.size):
        t = (exp[i] - calc[i]) ** 2 / exp[i]
        if not (np.isnan(t) or np.isinf(t)):
            sm1 += t
            sm2 += abs(exp[i])
    try:
        return sqrt(sm1 / sm2) * 100
    except:
        return 0

def Rpe(exp, calc, num_params):
    """
        Calculate the 'expected pattern R factor' (Rpe) for the given experimental and
        calculated pattern and the number of refined parameters.
         
    """
    # R expected:
    # Re = Sqrt( (Points - Params) / Sum[ w * obs² ] )
    num_points = exp.size
    return np.sqrt((num_points - num_params) / np.sum(exp ** 2)) * 100

def Rphase(exp, calc, phase):
    """
        Calculates the 'weighted phase R factor' (aka Rphase) for the given experimental and
        calculated patterns (1D numpy arrays, y-values only)
        The weights are set equal to the inverse of the observed intensities. 
    """
    # weighted Rp:
    # Rwp = Sqrt ( Sum[w * (obs - calc)²] / Sum[w * obs²] )  w = 1 / Iobs
    sm1 = 0
    sm2 = 0
    exp = exp.flatten()
    calc = calc.flatten()
    phase = phase.flatten()   
    for i in range(exp.size):
        t = (exp[i] - calc[i]) ** 2 * phase[i] / (exp[i]**2)
        if not (np.isnan(t) or np.isinf(t)):
            sm1 += t
            sm2 += abs(phase[i])
    try:
        return sqrt(sm1 / sm2) * 100
    except:
        return 0
    
